from __future__ import annotations

from typing import Optional

from fontTools.annotations import KerningPair, KerningDict, KerningGroups, IntFloat

StrDict = dict[str, str]


def lookupKerningValue(
    pair: KerningPair,
    kerning: KerningDict,
    groups: KerningGroups,
    fallback: IntFloat = 0,
    glyphToFirstGroup: Optional[StrDict] = None,
    glyphToSecondGroup: Optional[StrDict] = None,
) -> IntFloat:
    """Retrieve the kerning value (if any) between a pair of elements.

    The elments can be either individual glyphs (by name) or kerning
    groups (by name), or any combination of the two.

    Args:
      pair:
          A tuple, in logical order (first, second) with respect
          to the reading direction, to query the font for kerning
          information on. Each element in the tuple can be either
          a glyph name or a kerning group name.
      kerning:
          A dictionary of kerning pairs.
      groups:
          A set of kerning groups.
      fallback:
          The fallback value to return if no kern is found between
          the elements in ``pair``. Defaults to 0.
      glyphToFirstGroup:
          A dictionary mapping glyph names to the first-glyph kerning
          groups to which they belong. Defaults to ``None``.
      glyphToSecondGroup:
          A dictionary mapping glyph names to the second-glyph kerning
          groups to which they belong. Defaults to ``None``.

    Returns:
      The kerning value between the element pair. If no kerning for
      the pair is found, the fallback value is returned.

    Note: This function expects the ``kerning`` argument to be a flat
    dictionary of kerning pairs, not the nested structure used in a
    kerning.plist file.

    Examples::

      >>> groups = {
      ...     "public.kern1.O" : ["O", "D", "Q"],
      ...     "public.kern2.E" : ["E", "F"]
      ... }
      >>> kerning = {
      ...     ("public.kern1.O", "public.kern2.E") : -100,
      ...     ("public.kern1.O", "F") : -200,
      ...     ("D", "F") : -300
      ... }
      >>> lookupKerningValue(("D", "F"), kerning, groups)
      -300
      >>> lookupKerningValue(("O", "F"), kerning, groups)
      -200
      >>> lookupKerningValue(("O", "E"), kerning, groups)
      -100
      >>> lookupKerningValue(("O", "O"), kerning, groups)
      0
      >>> lookupKerningValue(("E", "E"), kerning, groups)
      0
      >>> lookupKerningValue(("E", "O"), kerning, groups)
      0
      >>> lookupKerningValue(("X", "X"), kerning, groups)
      0
      >>> lookupKerningValue(("public.kern1.O", "public.kern2.E"),
      ...     kerning, groups)
      -100
      >>> lookupKerningValue(("public.kern1.O", "F"), kerning, groups)
      -200
      >>> lookupKerningValue(("O", "public.kern2.E"), kerning, groups)
      -100
      >>> lookupKerningValue(("public.kern1.X", "public.kern2.X"), kerning, groups)
      0
    """
    # quickly check to see if the pair is in the kerning dictionary
    if pair in kerning:
        return kerning[pair]
    # ensure both or no glyph-to-group mappings are provided
    if (glyphToFirstGroup is None) != (glyphToSecondGroup is None):
        raise ValueError(
            "Must provide both 'glyphToFirstGroup' and 'glyphToSecondGroup', or neither."
        )
    # create glyph to group mapping
    if glyphToFirstGroup is None:
        glyphToFirstGroup = {}
        glyphToSecondGroup = {}
        for group, groupMembers in groups.items():
            if group.startswith("public.kern1."):
                for glyph in groupMembers:
                    glyphToFirstGroup[glyph] = group
            elif group.startswith("public.kern2."):
                for glyph in groupMembers:
                    glyphToSecondGroup[glyph] = group
    # ensure type safety for mappings
    assert glyphToFirstGroup is not None
    assert glyphToSecondGroup is not None
    # get group names and make sure first and second are glyph names
    first, second = pair
    firstGroup = secondGroup = None
    if first.startswith("public.kern1."):
        firstGroup = first
        firstGlyph = None
    else:
        firstGroup = glyphToFirstGroup.get(first)
        firstGlyph = first
    if second.startswith("public.kern2."):
        secondGroup = second
        secondGlyph = None
    else:
        secondGroup = glyphToSecondGroup.get(second)
        secondGlyph = second
    # make an ordered list of pairs to look up
    pairs = [
        (a, b)
        for a in (firstGlyph, firstGroup)
        for b in (secondGlyph, secondGroup)
        if a is not None and b is not None
    ]
    # look up the pairs and return any matches
    for pair in pairs:
        if pair in kerning:
            return kerning[pair]
    # use the fallback value
    return fallback


if __name__ == "__main__":
    import doctest

    doctest.testmod()



"""Module for reading and writing AFM (Adobe Font Metrics) files.

Note that this has been designed to read in AFM files generated by Fontographer
and has not been tested on many other files. In particular, it does not
implement the whole Adobe AFM specification [#f1]_ but, it should read most
"common" AFM files.

Here is an example of using `afmLib` to read, modify and write an AFM file:

	>>> from fontTools.afmLib import AFM
	>>> f = AFM("Tests/afmLib/data/TestAFM.afm")
	>>>
	>>> # Accessing a pair gets you the kern value
	>>> f[("V","A")]
	-60
	>>>
	>>> # Accessing a glyph name gets you metrics
	>>> f["A"]
	(65, 668, (8, -25, 660, 666))
	>>> # (charnum, width, bounding box)
	>>>
	>>> # Accessing an attribute gets you metadata
	>>> f.FontName
	'TestFont-Regular'
	>>> f.FamilyName
	'TestFont'
	>>> f.Weight
	'Regular'
	>>> f.XHeight
	500
	>>> f.Ascender
	750
	>>>
	>>> # Attributes and items can also be set
	>>> f[("A","V")] = -150 # Tighten kerning
	>>> f.FontName = "TestFont Squished"
	>>>
	>>> # And the font written out again (remove the # in front)
	>>> #f.write("testfont-squished.afm")

.. rubric:: Footnotes

.. [#f1] `Adobe Technote 5004 <https://www.adobe.com/content/dam/acom/en/devnet/font/pdfs/5004.AFM_Spec.pdf>`_,
   Adobe Font Metrics File Format Specification.

"""

import re

# every single line starts with a "word"
identifierRE = re.compile(r"^([A-Za-z]+).*")

# regular expression to parse char lines
charRE = re.compile(
    r"(-?\d+)"  # charnum
    r"\s*;\s*WX\s+"  # ; WX
    r"(-?\d+)"  # width
    r"\s*;\s*N\s+"  # ; N
    r"([.A-Za-z0-9_]+)"  # charname
    r"\s*;\s*B\s+"  # ; B
    r"(-?\d+)"  # left
    r"\s+"
    r"(-?\d+)"  # bottom
    r"\s+"
    r"(-?\d+)"  # right
    r"\s+"
    r"(-?\d+)"  # top
    r"\s*;\s*"  # ;
)

# regular expression to parse kerning lines
kernRE = re.compile(
    r"([.A-Za-z0-9_]+)"  # leftchar
    r"\s+"
    r"([.A-Za-z0-9_]+)"  # rightchar
    r"\s+"
    r"(-?\d+)"  # value
    r"\s*"
)

# regular expressions to parse composite info lines of the form:
# Aacute 2 ; PCC A 0 0 ; PCC acute 182 211 ;
compositeRE = re.compile(
    r"([.A-Za-z0-9_]+)"  # char name
    r"\s+"
    r"(\d+)"  # number of parts
    r"\s*;\s*"
)
componentRE = re.compile(
    r"PCC\s+"  # PPC
    r"([.A-Za-z0-9_]+)"  # base char name
    r"\s+"
    r"(-?\d+)"  # x offset
    r"\s+"
    r"(-?\d+)"  # y offset
    r"\s*;\s*"
)

preferredAttributeOrder = [
    "FontName",
    "FullName",
    "FamilyName",
    "Weight",
    "ItalicAngle",
    "IsFixedPitch",
    "FontBBox",
    "UnderlinePosition",
    "UnderlineThickness",
    "Version",
    "Notice",
    "EncodingScheme",
    "CapHeight",
    "XHeight",
    "Ascender",
    "Descender",
]


class error(Exception):
    pass


class AFM(object):
    _attrs = None

    _keywords = [
        "StartFontMetrics",
        "EndFontMetrics",
        "StartCharMetrics",
        "EndCharMetrics",
        "StartKernData",
        "StartKernPairs",
        "EndKernPairs",
        "EndKernData",
        "StartComposites",
        "EndComposites",
    ]

    def __init__(self, path=None):
        """AFM file reader.

        Instantiating an object with a path name will cause the file to be opened,
        read, and parsed. Alternatively the path can be left unspecified, and a
        file can be parsed later with the :meth:`read` method."""
        self._attrs = {}
        self._chars = {}
        self._kerning = {}
        self._index = {}
        self._comments = []
        self._composites = {}
        if path is not None:
            self.read(path)

    def read(self, path):
        """Opens, reads and parses a file."""
        lines = readlines(path)
        for line in lines:
            if not line.strip():
                continue
            m = identifierRE.match(line)
            if m is None:
                raise error("syntax error in AFM file: " + repr(line))

            pos = m.regs[1][1]
            word = line[:pos]
            rest = line[pos:].strip()
            if word in self._keywords:
                continue
            if word == "C":
                self.parsechar(rest)
            elif word == "KPX":
                self.parsekernpair(rest)
            elif word == "CC":
                self.parsecomposite(rest)
            else:
                self.parseattr(word, rest)

    def parsechar(self, rest):
        m = charRE.match(rest)
        if m is None:
            raise error("syntax error in AFM file: " + repr(rest))
        things = []
        for fr, to in m.regs[1:]:
            things.append(rest[fr:to])
        charname = things[2]
        del things[2]
        charnum, width, l, b, r, t = (int(thing) for thing in things)
        self._chars[charname] = charnum, width, (l, b, r, t)

    def parsekernpair(self, rest):
        m = kernRE.match(rest)
        if m is None:
            raise error("syntax error in AFM file: " + repr(rest))
        things = []
        for fr, to in m.regs[1:]:
            things.append(rest[fr:to])
        leftchar, rightchar, value = things
        value = int(value)
        self._kerning[(leftchar, rightchar)] = value

    def parseattr(self, word, rest):
        if word == "FontBBox":
            l, b, r, t = [int(thing) for thing in rest.split()]
            self._attrs[word] = l, b, r, t
        elif word == "Comment":
            self._comments.append(rest)
        else:
            try:
                value = int(rest)
            except (ValueError, OverflowError):
                self._attrs[word] = rest
            else:
                self._attrs[word] = value

    def parsecomposite(self, rest):
        m = compositeRE.match(rest)
        if m is None:
            raise error("syntax error in AFM file: " + repr(rest))
        charname = m.group(1)
        ncomponents = int(m.group(2))
        rest = rest[m.regs[0][1] :]
        components = []
        while True:
            m = componentRE.match(rest)
            if m is None:
                raise error("syntax error in AFM file: " + repr(rest))
            basechar = m.group(1)
            xoffset = int(m.group(2))
            yoffset = int(m.group(3))
            components.append((basechar, xoffset, yoffset))
            rest = rest[m.regs[0][1] :]
            if not rest:
                break
        assert len(components) == ncomponents
        self._composites[charname] = components

    def write(self, path, sep="\r"):
        """Writes out an AFM font to the given path."""
        import time

        lines = [
            "StartFontMetrics 2.0",
            "Comment Generated by afmLib; at %s"
            % (time.strftime("%m/%d/%Y %H:%M:%S", time.localtime(time.time()))),
        ]

        # write comments, assuming (possibly wrongly!) they should
        # all appear at the top
        for comment in self._comments:
            lines.append("Comment " + comment)

        # write attributes, first the ones we know about, in
        # a preferred order
        attrs = self._attrs
        for attr in preferredAttributeOrder:
            if attr in attrs:
                value = attrs[attr]
                if attr == "FontBBox":
                    value = "%s %s %s %s" % value
                lines.append(attr + " " + str(value))
        # then write the attributes we don't know about,
        # in alphabetical order
        items = sorted(attrs.items())
        for attr, value in items:
            if attr in preferredAttributeOrder:
                continue
            lines.append(attr + " " + str(value))

        # write char metrics
        lines.append("StartCharMetrics " + repr(len(self._chars)))
        items = [
            (charnum, (charname, width, box))
            for charname, (charnum, width, box) in self._chars.items()
        ]

        def myKey(a):
            """Custom key function to make sure unencoded chars (-1)
            end up at the end of the list after sorting."""
            if a[0] == -1:
                a = (0xFFFF,) + a[1:]  # 0xffff is an arbitrary large number
            return a

        items.sort(key=myKey)

        for charnum, (charname, width, (l, b, r, t)) in items:
            lines.append(
                "C %d ; WX %d ; N %s ; B %d %d %d %d ;"
                % (charnum, width, charname, l, b, r, t)
            )
        lines.append("EndCharMetrics")

        # write kerning info
        lines.append("StartKernData")
        lines.append("StartKernPairs " + repr(len(self._kerning)))
        items = sorted(self._kerning.items())
        for (leftchar, rightchar), value in items:
            lines.append("KPX %s %s %d" % (leftchar, rightchar, value))
        lines.append("EndKernPairs")
        lines.append("EndKernData")

        if self._composites:
            composites = sorted(self._composites.items())
            lines.append("StartComposites %s" % len(self._composites))
            for charname, components in composites:
                line = "CC %s %s ;" % (charname, len(components))
                for basechar, xoffset, yoffset in components:
                    line = line + " PCC %s %s %s ;" % (basechar, xoffset, yoffset)
                lines.append(line)
            lines.append("EndComposites")

        lines.append("EndFontMetrics")

        writelines(path, lines, sep)

    def has_kernpair(self, pair):
        """Returns `True` if the given glyph pair (specified as a tuple) exists
        in the kerning dictionary."""
        return pair in self._kerning

    def kernpairs(self):
        """Returns a list of all kern pairs in the kerning dictionary."""
        return list(self._kerning.keys())

    def has_char(self, char):
        """Returns `True` if the given glyph exists in the font."""
        return char in self._chars

    def chars(self):
        """Returns a list of all glyph names in the font."""
        return list(self._chars.keys())

    def comments(self):
        """Returns all comments from the file."""
        return self._comments

    def addComment(self, comment):
        """Adds a new comment to the file."""
        self._comments.append(comment)

    def addComposite(self, glyphName, components):
        """Specifies that the glyph `glyphName` is made up of the given components.
        The components list should be of the following form::

                [
                        (glyphname, xOffset, yOffset),
                        ...
                ]

        """
        self._composites[glyphName] = components

    def __getattr__(self, attr):
        if attr in self._attrs:
            return self._attrs[attr]
        else:
            raise AttributeError(attr)

    def __setattr__(self, attr, value):
        # all attrs *not* starting with "_" are consider to be AFM keywords
        if attr[:1] == "_":
            self.__dict__[attr] = value
        else:
            self._attrs[attr] = value

    def __delattr__(self, attr):
        # all attrs *not* starting with "_" are consider to be AFM keywords
        if attr[:1] == "_":
            try:
                del self.__dict__[attr]
            except KeyError:
                raise AttributeError(attr)
        else:
            try:
                del self._attrs[attr]
            except KeyError:
                raise AttributeError(attr)

    def __getitem__(self, key):
        if isinstance(key, tuple):
            # key is a tuple, return the kernpair
            return self._kerning[key]
        else:
            # return the metrics instead
            return self._chars[key]

    def __setitem__(self, key, value):
        if isinstance(key, tuple):
            # key is a tuple, set kernpair
            self._kerning[key] = value
        else:
            # set char metrics
            self._chars[key] = value

    def __delitem__(self, key):
        if isinstance(key, tuple):
            # key is a tuple, del kernpair
            del self._kerning[key]
        else:
            # del char metrics
            del self._chars[key]

    def __repr__(self):
        if hasattr(self, "FullName"):
            return "<AFM object for %s>" % self.FullName
        else:
            return "<AFM object at %x>" % id(self)


def readlines(path):
    with open(path, "r", encoding="ascii") as f:
        data = f.read()
    return data.splitlines()


def writelines(path, lines, sep="\r"):
    with open(path, "w", encoding="ascii", newline=sep) as f:
        f.write("\n".join(lines) + "\n")


if __name__ == "__main__":
    import EasyDialogs

    path = EasyDialogs.AskFileForOpen()
    if path:
        afm = AFM(path)
        char = "A"
        if afm.has_char(char):
            print(afm[char])  # print charnum, width and boundingbox
        pair = ("A", "V")
        if afm.has_kernpair(pair):
            print(afm[pair])  # print kerning value for pair
        print(afm.Version)  # various other afm entries have become attributes
        print(afm.Weight)
        # afm.comments() returns a list of all Comment lines found in the AFM
        print(afm.comments())
        # print afm.chars()
        # print afm.kernpairs()
        print(afm)
        afm.write(path + ".muck")



import fontTools.misc.testTools as testTools
import unittest


class TestToolsTest(unittest.TestCase):
    def test_parseXML_str(self):
        self.assertEqual(
            testTools.parseXML(
                '<Foo n="1"/>'
                '<Foo n="2">'
                "    some ünıcòðe text"
                '    <Bar color="red"/>'
                "    some more text"
                "</Foo>"
                '<Foo n="3"/>'
            ),
            [
                ("Foo", {"n": "1"}, []),
                (
                    "Foo",
                    {"n": "2"},
                    [
                        "    some ünıcòðe text    ",
                        ("Bar", {"color": "red"}, []),
                        "    some more text",
                    ],
                ),
                ("Foo", {"n": "3"}, []),
            ],
        )

    def test_parseXML_bytes(self):
        self.assertEqual(
            testTools.parseXML(
                b'<Foo n="1"/>'
                b'<Foo n="2">'
                b"    some \xc3\xbcn\xc4\xb1c\xc3\xb2\xc3\xb0e text"
                b'    <Bar color="red"/>'
                b"    some more text"
                b"</Foo>"
                b'<Foo n="3"/>'
            ),
            [
                ("Foo", {"n": "1"}, []),
                (
                    "Foo",
                    {"n": "2"},
                    [
                        "    some ünıcòðe text    ",
                        ("Bar", {"color": "red"}, []),
                        "    some more text",
                    ],
                ),
                ("Foo", {"n": "3"}, []),
            ],
        )

    def test_parseXML_str_list(self):
        self.assertEqual(
            testTools.parseXML(['<Foo n="1"/>' '<Foo n="2"/>']),
            [("Foo", {"n": "1"}, []), ("Foo", {"n": "2"}, [])],
        )

    def test_parseXML_bytes_list(self):
        self.assertEqual(
            testTools.parseXML([b'<Foo n="1"/>' b'<Foo n="2"/>']),
            [("Foo", {"n": "1"}, []), ("Foo", {"n": "2"}, [])],
        )

    def test_getXML(self):
        def toXML(writer, ttFont):
            writer.simpletag("simple")
            writer.newline()
            writer.begintag("tag", attr="value")
            writer.newline()
            writer.write("hello world")
            writer.newline()
            writer.endtag("tag")
            writer.newline()  # toXML always ends with a newline

        self.assertEqual(
            testTools.getXML(toXML),
            ["<simple/>", '<tag attr="value">', "  hello world", "</tag>"],
        )


if __name__ == "__main__":
    import sys

    sys.exit(unittest.main())



import os
import shutil
import unittest
import tempfile
from io import open
from fontTools.ufoLib import UFOReader, UFOWriter, UFOLibError
from fontTools.ufoLib import plistlib
from .testSupport import fontInfoVersion1, fontInfoVersion2


class TestInfoObject:
    pass


class ReadFontInfoVersion1TestCase(unittest.TestCase):
    def setUp(self):
        self.dstDir = tempfile.mktemp()
        os.mkdir(self.dstDir)
        metaInfo = {"creator": "test", "formatVersion": 1}
        path = os.path.join(self.dstDir, "metainfo.plist")
        with open(path, "wb") as f:
            plistlib.dump(metaInfo, f)

    def tearDown(self):
        shutil.rmtree(self.dstDir)

    def _writeInfoToPlist(self, info):
        path = os.path.join(self.dstDir, "fontinfo.plist")
        with open(path, "wb") as f:
            plistlib.dump(info, f)

    def testRead(self):
        originalData = dict(fontInfoVersion1)
        self._writeInfoToPlist(originalData)
        infoObject = TestInfoObject()
        reader = UFOReader(self.dstDir, validate=True)
        reader.readInfo(infoObject)
        for attr in dir(infoObject):
            if attr not in fontInfoVersion2:
                continue
            originalValue = fontInfoVersion2[attr]
            readValue = getattr(infoObject, attr)
            self.assertEqual(originalValue, readValue)

    def testFontStyleConversion(self):
        fontStyle1To2 = {64: "regular", 1: "italic", 32: "bold", 33: "bold italic"}
        for old, new in list(fontStyle1To2.items()):
            info = dict(fontInfoVersion1)
            info["fontStyle"] = old
            self._writeInfoToPlist(info)
            reader = UFOReader(self.dstDir, validate=True)
            infoObject = TestInfoObject()
            reader.readInfo(infoObject)
            self.assertEqual(new, infoObject.styleMapStyleName)

    def testWidthNameConversion(self):
        widthName1To2 = {
            "Ultra-condensed": 1,
            "Extra-condensed": 2,
            "Condensed": 3,
            "Semi-condensed": 4,
            "Medium (normal)": 5,
            "Semi-expanded": 6,
            "Expanded": 7,
            "Extra-expanded": 8,
            "Ultra-expanded": 9,
        }
        for old, new in list(widthName1To2.items()):
            info = dict(fontInfoVersion1)
            info["widthName"] = old
            self._writeInfoToPlist(info)
            reader = UFOReader(self.dstDir, validate=True)
            infoObject = TestInfoObject()
            reader.readInfo(infoObject)
            self.assertEqual(new, infoObject.openTypeOS2WidthClass)


class WriteFontInfoVersion1TestCase(unittest.TestCase):
    def setUp(self):
        self.tempDir = tempfile.mktemp()
        os.mkdir(self.tempDir)
        self.dstDir = os.path.join(self.tempDir, "test.ufo")

    def tearDown(self):
        shutil.rmtree(self.tempDir)

    def makeInfoObject(self):
        infoObject = TestInfoObject()
        for attr, value in list(fontInfoVersion2.items()):
            setattr(infoObject, attr, value)
        return infoObject

    def readPlist(self):
        path = os.path.join(self.dstDir, "fontinfo.plist")
        with open(path, "rb") as f:
            plist = plistlib.load(f)
        return plist

    def testWrite(self):
        infoObject = self.makeInfoObject()
        writer = UFOWriter(self.dstDir, formatVersion=1)
        writer.writeInfo(infoObject)
        writtenData = self.readPlist()
        for attr, originalValue in list(fontInfoVersion1.items()):
            newValue = writtenData[attr]
            self.assertEqual(newValue, originalValue)

    def testFontStyleConversion(self):
        fontStyle1To2 = {64: "regular", 1: "italic", 32: "bold", 33: "bold italic"}
        for old, new in list(fontStyle1To2.items()):
            infoObject = self.makeInfoObject()
            infoObject.styleMapStyleName = new
            writer = UFOWriter(self.dstDir, formatVersion=1)
            writer.writeInfo(infoObject)
            writtenData = self.readPlist()
            self.assertEqual(writtenData["fontStyle"], old)

    def testWidthNameConversion(self):
        widthName1To2 = {
            "Ultra-condensed": 1,
            "Extra-condensed": 2,
            "Condensed": 3,
            "Semi-condensed": 4,
            "Medium (normal)": 5,
            "Semi-expanded": 6,
            "Expanded": 7,
            "Extra-expanded": 8,
            "Ultra-expanded": 9,
        }
        for old, new in list(widthName1To2.items()):
            infoObject = self.makeInfoObject()
            infoObject.openTypeOS2WidthClass = new
            writer = UFOWriter(self.dstDir, formatVersion=1)
            writer.writeInfo(infoObject)
            writtenData = self.readPlist()
            self.assertEqual(writtenData["widthName"], old)



#! /usr/bin/env python3

# Example script to decompose the composite glyphs in a TTF into
# non-composite outlines.


import sys
from fontTools.ttLib import TTFont
from fontTools.pens.recordingPen import DecomposingRecordingPen
from fontTools.pens.ttGlyphPen import TTGlyphPen

try:
    import pathops
except ImportError:
    sys.exit(
        "This script requires the skia-pathops module. "
        "`pip install skia-pathops` and then retry."
    )


if len(sys.argv) != 3:
    print("usage: decompose-ttf.py fontfile.ttf outfile.ttf")
    sys.exit(1)

src = sys.argv[1]
dst = sys.argv[2]

with TTFont(src) as f:
    glyfTable = f["glyf"]
    glyphSet = f.getGlyphSet()

    for glyphName in glyphSet.keys():
        if not glyfTable[glyphName].isComposite():
            continue

        # record TTGlyph outlines without components
        dcPen = DecomposingRecordingPen(glyphSet)
        glyphSet[glyphName].draw(dcPen)

        # replay recording onto a skia-pathops Path
        path = pathops.Path()
        pathPen = path.getPen()
        dcPen.replay(pathPen)

        # remove overlaps
        path.simplify()

        # create new TTGlyph from Path
        ttPen = TTGlyphPen(None)
        path.draw(ttPen)
        glyfTable[glyphName] = ttPen.glyph()

    f.save(dst)



from fontTools.misc.testTools import FakeFont, getXML, parseXML
from fontTools.misc.textTools import bytechr, bytesjoin, deHexStr, hexStr
from fontTools.ttLib import newTable
import unittest


# A simple 'morx' table with non-contextual glyph substitution.
# Unfortunately, the Apple spec for 'morx' does not contain a complete example.
# The test case has therefore been adapted from the example 'mort' table in
# https://developer.apple.com/fonts/TrueType-Reference-Manual/RM06/Chap6mort.html
MORX_NONCONTEXTUAL_DATA = deHexStr(
    "0002 0000 "  #  0: Version=2, Reserved=0
    "0000 0001 "  #  4: MorphChainCount=1
    "0000 0001 "  #  8: DefaultFlags=1
    "0000 0058 "  # 12: StructLength=88
    "0000 0003 "  # 16: MorphFeatureCount=3
    "0000 0001 "  # 20: MorphSubtableCount=1
    "0004 0000 "  # 24: Feature[0].FeatureType=4/VertSubst, .FeatureSetting=on
    "0000 0001 "  # 28: Feature[0].EnableFlags=0x00000001
    "FFFF FFFF "  # 32: Feature[0].DisableFlags=0xFFFFFFFF
    "0004 0001 "  # 36: Feature[1].FeatureType=4/VertSubst, .FeatureSetting=off
    "0000 0000 "  # 40: Feature[1].EnableFlags=0x00000000
    "FFFF FFFE "  # 44: Feature[1].DisableFlags=0xFFFFFFFE
    "0000 0001 "  # 48: Feature[2].FeatureType=0/GlyphEffects, .FeatSetting=off
    "0000 0000 "  # 52: Feature[2].EnableFlags=0 (required for last feature)
    "0000 0000 "  # 56: Feature[2].EnableFlags=0 (required for last feature)
    "0000 0024 "  # 60: Subtable[0].StructLength=36
    "80 "  # 64: Subtable[0].CoverageFlags=0x80
    "00 00 "  # 65: Subtable[0].Reserved=0
    "04 "  # 67: Subtable[0].MorphType=4/NoncontextualMorph
    "0000 0001 "  # 68: Subtable[0].SubFeatureFlags=0x1
    "0006 0004 "  # 72: LookupFormat=6, UnitSize=4
    "0002 0008 "  # 76: NUnits=2, SearchRange=8
    "0001 0000 "  # 80: EntrySelector=1, RangeShift=0
    "000B 0087 "  # 84: Glyph=11 (parenleft); Value=135 (parenleft.vertical)
    "000D 0088 "  # 88: Glyph=13 (parenright); Value=136 (parenright.vertical)
    "FFFF 0000 "  # 92: Glyph=<end>; Value=0
)  # 96: <end>
assert len(MORX_NONCONTEXTUAL_DATA) == 96


MORX_NONCONTEXTUAL_XML = [
    '<Version value="2"/>',
    '<Reserved value="0"/>',
    "<!-- MorphChainCount=1 -->",
    '<MorphChain index="0">',
    '  <DefaultFlags value="0x00000001"/>',
    "  <!-- StructLength=88 -->",
    "  <!-- MorphFeatureCount=3 -->",
    "  <!-- MorphSubtableCount=1 -->",
    '  <MorphFeature index="0">',
    '    <FeatureType value="4"/>',
    '    <FeatureSetting value="0"/>',
    '    <EnableFlags value="0x00000001"/>',
    '    <DisableFlags value="0xFFFFFFFF"/>',
    "  </MorphFeature>",
    '  <MorphFeature index="1">',
    '    <FeatureType value="4"/>',
    '    <FeatureSetting value="1"/>',
    '    <EnableFlags value="0x00000000"/>',
    '    <DisableFlags value="0xFFFFFFFE"/>',
    "  </MorphFeature>",
    '  <MorphFeature index="2">',
    '    <FeatureType value="0"/>',
    '    <FeatureSetting value="1"/>',
    '    <EnableFlags value="0x00000000"/>',
    '    <DisableFlags value="0x00000000"/>',
    "  </MorphFeature>",
    '  <MorphSubtable index="0">',
    "    <!-- StructLength=36 -->",
    '    <TextDirection value="Vertical"/>',
    '    <ProcessingOrder value="LayoutOrder"/>',
    "    <!-- MorphType=4 -->",
    '    <SubFeatureFlags value="0x00000001"/>',
    "    <NoncontextualMorph>",
    "      <Substitution>",
    '        <Lookup glyph="parenleft" value="parenleft.vertical"/>',
    '        <Lookup glyph="parenright" value="parenright.vertical"/>',
    "      </Substitution>",
    "    </NoncontextualMorph>",
    "  </MorphSubtable>",
    "</MorphChain>",
]


MORX_REARRANGEMENT_DATA = deHexStr(
    "0002 0000 "  #  0: Version=2, Reserved=0
    "0000 0001 "  #  4: MorphChainCount=1
    "0000 0001 "  #  8: DefaultFlags=1
    "0000 0078 "  # 12: StructLength=120 (+8=128)
    "0000 0000 "  # 16: MorphFeatureCount=0
    "0000 0001 "  # 20: MorphSubtableCount=1
    "0000 0068 "  # 24: Subtable[0].StructLength=104 (+24=128)
    "80 "  # 28: Subtable[0].CoverageFlags=0x80
    "00 00 "  # 29: Subtable[0].Reserved=0
    "00 "  # 31: Subtable[0].MorphType=0/RearrangementMorph
    "0000 0001 "  # 32: Subtable[0].SubFeatureFlags=0x1
    "0000 0006 "  # 36: STXHeader.ClassCount=6
    "0000 0010 "  # 40: STXHeader.ClassTableOffset=16 (+36=52)
    "0000 0028 "  # 44: STXHeader.StateArrayOffset=40 (+36=76)
    "0000 004C "  # 48: STXHeader.EntryTableOffset=76 (+36=112)
    "0006 0004 "  # 52: ClassTable.LookupFormat=6, .UnitSize=4
    "0002 0008 "  # 56:   .NUnits=2, .SearchRange=8
    "0001 0000 "  # 60:   .EntrySelector=1, .RangeShift=0
    "0001 0005 "  # 64:   Glyph=A; Class=5
    "0003 0004 "  # 68:   Glyph=C; Class=4
    "FFFF 0000 "  # 72:   Glyph=<end>; Value=0
    "0000 0001 0002 0003 0002 0001 "  #  76: State[0][0..5]
    "0003 0003 0003 0003 0003 0003 "  #  88: State[1][0..5]
    "0001 0003 0003 0003 0002 0002 "  # 100: State[2][0..5]
    "0002 FFFF "  # 112: Entries[0].NewState=2, .Flags=0xFFFF
    "0001 A00D "  # 116: Entries[1].NewState=1, .Flags=0xA00D
    "0000 8006 "  # 120: Entries[2].NewState=0, .Flags=0x8006
    "0002 0000 "  # 124: Entries[3].NewState=2, .Flags=0x0000
)  # 128: <end>
assert len(MORX_REARRANGEMENT_DATA) == 128, len(MORX_REARRANGEMENT_DATA)


MORX_REARRANGEMENT_XML = [
    '<Version value="2"/>',
    '<Reserved value="0"/>',
    "<!-- MorphChainCount=1 -->",
    '<MorphChain index="0">',
    '  <DefaultFlags value="0x00000001"/>',
    "  <!-- StructLength=120 -->",
    "  <!-- MorphFeatureCount=0 -->",
    "  <!-- MorphSubtableCount=1 -->",
    '  <MorphSubtable index="0">',
    "    <!-- StructLength=104 -->",
    '    <TextDirection value="Vertical"/>',
    '    <ProcessingOrder value="LayoutOrder"/>',
    "    <!-- MorphType=0 -->",
    '    <SubFeatureFlags value="0x00000001"/>',
    "    <RearrangementMorph>",
    "      <StateTable>",
    "        <!-- GlyphClassCount=6 -->",
    '        <GlyphClass glyph="A" value="5"/>',
    '        <GlyphClass glyph="C" value="4"/>',
    '        <State index="0">',
    '          <Transition onGlyphClass="0">',
    '            <NewState value="2"/>',
    '            <Flags value="MarkFirst,DontAdvance,MarkLast"/>',
    '            <ReservedFlags value="0x1FF0"/>',
    '            <Verb value="15"/><!-- ABxCD ⇒ DCxBA -->',
    "          </Transition>",
    '          <Transition onGlyphClass="1">',
    '            <NewState value="1"/>',
    '            <Flags value="MarkFirst,MarkLast"/>',
    '            <Verb value="13"/><!-- ABxCD ⇒ CDxBA -->',
    "          </Transition>",
    '          <Transition onGlyphClass="2">',
    '            <NewState value="0"/>',
    '            <Flags value="MarkFirst"/>',
    '            <Verb value="6"/><!-- xCD ⇒ CDx -->',
    "          </Transition>",
    '          <Transition onGlyphClass="3">',
    '            <NewState value="2"/>',
    '            <Verb value="0"/><!-- no change -->',
    "          </Transition>",
    '          <Transition onGlyphClass="4">',
    '            <NewState value="0"/>',
    '            <Flags value="MarkFirst"/>',
    '            <Verb value="6"/><!-- xCD ⇒ CDx -->',
    "          </Transition>",
    '          <Transition onGlyphClass="5">',
    '            <NewState value="1"/>',
    '            <Flags value="MarkFirst,MarkLast"/>',
    '            <Verb value="13"/><!-- ABxCD ⇒ CDxBA -->',
    "          </Transition>",
    "        </State>",
    '        <State index="1">',
    '          <Transition onGlyphClass="0">',
    '            <NewState value="2"/>',
    '            <Verb value="0"/><!-- no change -->',
    "          </Transition>",
    '          <Transition onGlyphClass="1">',
    '            <NewState value="2"/>',
    '            <Verb value="0"/><!-- no change -->',
    "          </Transition>",
    '          <Transition onGlyphClass="2">',
    '            <NewState value="2"/>',
    '            <Verb value="0"/><!-- no change -->',
    "          </Transition>",
    '          <Transition onGlyphClass="3">',
    '            <NewState value="2"/>',
    '            <Verb value="0"/><!-- no change -->',
    "          </Transition>",
    '          <Transition onGlyphClass="4">',
    '            <NewState value="2"/>',
    '            <Verb value="0"/><!-- no change -->',
    "          </Transition>",
    '          <Transition onGlyphClass="5">',
    '            <NewState value="2"/>',
    '            <Verb value="0"/><!-- no change -->',
    "          </Transition>",
    "        </State>",
    '        <State index="2">',
    '          <Transition onGlyphClass="0">',
    '            <NewState value="1"/>',
    '            <Flags value="MarkFirst,MarkLast"/>',
    '            <Verb value="13"/><!-- ABxCD ⇒ CDxBA -->',
    "          </Transition>",
    '          <Transition onGlyphClass="1">',
    '            <NewState value="2"/>',
    '            <Verb value="0"/><!-- no change -->',
    "          </Transition>",
    '          <Transition onGlyphClass="2">',
    '            <NewState value="2"/>',
    '            <Verb value="0"/><!-- no change -->',
    "          </Transition>",
    '          <Transition onGlyphClass="3">',
    '            <NewState value="2"/>',
    '            <Verb value="0"/><!-- no change -->',
    "          </Transition>",
    '          <Transition onGlyphClass="4">',
    '            <NewState value="0"/>',
    '            <Flags value="MarkFirst"/>',
    '            <Verb value="6"/><!-- xCD ⇒ CDx -->',
    "          </Transition>",
    '          <Transition onGlyphClass="5">',
    '            <NewState value="0"/>',
    '            <Flags value="MarkFirst"/>',
    '            <Verb value="6"/><!-- xCD ⇒ CDx -->',
    "          </Transition>",
    "        </State>",
    "      </StateTable>",
    "    </RearrangementMorph>",
    "  </MorphSubtable>",
    "</MorphChain>",
]


# Taken from “Example 1: A contextal substituation table” in
# https://developer.apple.com/fonts/TrueType-Reference-Manual/RM06/Chap6morx.html
# as retrieved on 2017-09-05.
#
# Compared to the example table in Apple’s specification, we’ve
# made the following changes:
#
# * at offsets 0..35, we’ve prepended 36 bytes of boilerplate
#   to make the data a structurally valid ‘morx’ table;
#
# * at offset 36 (offset 0 in Apple’s document), we’ve changed
#   the number of glyph classes from 5 to 6 because the encoded
#   finite-state machine has transitions for six different glyph
#   classes (0..5);
#
# * at offset 52 (offset 16 in Apple’s document), we’ve replaced
#   the presumably leftover ‘XXX’ mark by an actual data offset;
#
# * at offset 72 (offset 36 in Apple’s document), we’ve changed
#   the input GlyphID from 51 to 52. With the original value of 51,
#   the glyph class lookup table can be encoded with equally many
#   bytes in either format 2 or 6; after changing the GlyphID to 52,
#   the most compact encoding is lookup format 6, as used in Apple’s
#   example;
#
# * at offset 90 (offset 54 in Apple’s document), we’ve changed
#   the value for the lookup end-of-table marker from 1 to 0.
#   Fonttools always uses zero for this value, whereas Apple’s
#   spec examples are inconsistently using one of {0, 1, 0xFFFF}
#   for this filler value;
#
# * at offset 172 (offset 136 in Apple’s document), we’ve again changed
#   the input GlyphID from 51 to 52, for the same reason as above.
#
# TODO: Ask Apple to fix “Example 1” in the ‘morx’ specification.
MORX_CONTEXTUAL_DATA = deHexStr(
    "0002 0000 "  #  0: Version=2, Reserved=0
    "0000 0001 "  #  4: MorphChainCount=1
    "0000 0001 "  #  8: DefaultFlags=1
    "0000 00B4 "  # 12: StructLength=180 (+8=188)
    "0000 0000 "  # 16: MorphFeatureCount=0
    "0000 0001 "  # 20: MorphSubtableCount=1
    "0000 00A4 "  # 24: Subtable[0].StructLength=164 (+24=188)
    "80 "  # 28: Subtable[0].CoverageFlags=0x80
    "00 00 "  # 29: Subtable[0].Reserved=0
    "01 "  # 31: Subtable[0].MorphType=1/ContextualMorph
    "0000 0001 "  # 32: Subtable[0].SubFeatureFlags=0x1
    "0000 0006 "  # 36: STXHeader.ClassCount=6
    "0000 0014 "  # 40: STXHeader.ClassTableOffset=20 (+36=56)
    "0000 0038 "  # 44: STXHeader.StateArrayOffset=56 (+36=92)
    "0000 005C "  # 48: STXHeader.EntryTableOffset=92 (+36=128)
    "0000 0074 "  # 52: STXHeader.PerGlyphTableOffset=116 (+36=152)
    # Glyph class table.
    "0006 0004 "  # 56: ClassTable.LookupFormat=6, .UnitSize=4
    "0005 0010 "  # 60:   .NUnits=5, .SearchRange=16
    "0002 0004 "  # 64:   .EntrySelector=2, .RangeShift=4
    "0032 0004 "  # 68:   Glyph=50; Class=4
    "0034 0004 "  # 72:   Glyph=52; Class=4
    "0050 0005 "  # 76:   Glyph=80; Class=5
    "00C9 0004 "  # 80:   Glyph=201; Class=4
    "00CA 0004 "  # 84:   Glyph=202; Class=4
    "FFFF 0000 "  # 88:   Glyph=<end>; Value=<filler>
    # State array.
    "0000 0000 0000 0000 0000 0001 "  #  92: State[0][0..5]
    "0000 0000 0000 0000 0000 0001 "  # 104: State[1][0..5]
    "0000 0000 0000 0000 0002 0001 "  # 116: State[2][0..5]
    # Entry table.
    "0000 0000 "  # 128: Entries[0].NewState=0, .Flags=0
    "FFFF FFFF "  # 132: Entries[0].MarkSubst=None, .CurSubst=None
    "0002 0000 "  # 136: Entries[1].NewState=2, .Flags=0
    "FFFF FFFF "  # 140: Entries[1].MarkSubst=None, .CurSubst=None
    "0000 0000 "  # 144: Entries[2].NewState=0, .Flags=0
    "FFFF 0000 "  # 148: Entries[2].MarkSubst=None, .CurSubst=PerGlyph #0
    # 152: <no padding needed for 4-byte alignment>
    # Per-glyph lookup tables.
    "0000 0004 "  # 152: Offset from this point to per-glyph lookup #0.
    # Per-glyph lookup #0.
    "0006 0004 "  # 156: ClassTable.LookupFormat=6, .UnitSize=4
    "0004 0010 "  # 160:   .NUnits=4, .SearchRange=16
    "0002 0000 "  # 164:   .EntrySelector=2, .RangeShift=0
    "0032 0258 "  # 168:   Glyph=50; ReplacementGlyph=600
    "0034 0259 "  # 172:   Glyph=52; ReplacementGlyph=601
    "00C9 025A "  # 176:   Glyph=201; ReplacementGlyph=602
    "00CA 0384 "  # 180:   Glyph=202; ReplacementGlyph=900
    "FFFF 0000 "  # 184:   Glyph=<end>; Value=<filler>
)  # 188: <end>
assert len(MORX_CONTEXTUAL_DATA) == 188, len(MORX_CONTEXTUAL_DATA)


MORX_CONTEXTUAL_XML = [
    '<Version value="2"/>',
    '<Reserved value="0"/>',
    "<!-- MorphChainCount=1 -->",
    '<MorphChain index="0">',
    '  <DefaultFlags value="0x00000001"/>',
    "  <!-- StructLength=180 -->",
    "  <!-- MorphFeatureCount=0 -->",
    "  <!-- MorphSubtableCount=1 -->",
    '  <MorphSubtable index="0">',
    "    <!-- StructLength=164 -->",
    '    <TextDirection value="Vertical"/>',
    '    <ProcessingOrder value="LayoutOrder"/>',
    "    <!-- MorphType=1 -->",
    '    <SubFeatureFlags value="0x00000001"/>',
    "    <ContextualMorph>",
    "      <StateTable>",
    "        <!-- GlyphClassCount=6 -->",
    '        <GlyphClass glyph="A" value="4"/>',
    '        <GlyphClass glyph="B" value="4"/>',
    '        <GlyphClass glyph="C" value="5"/>',
    '        <GlyphClass glyph="X" value="4"/>',
    '        <GlyphClass glyph="Y" value="4"/>',
    '        <State index="0">',
    '          <Transition onGlyphClass="0">',
    '            <NewState value="0"/>',
    '            <MarkIndex value="65535"/>',
    '            <CurrentIndex value="65535"/>',
    "          </Transition>",
    '          <Transition onGlyphClass="1">',
    '            <NewState value="0"/>',
    '            <MarkIndex value="65535"/>',
    '            <CurrentIndex value="65535"/>',
    "          </Transition>",
    '          <Transition onGlyphClass="2">',
    '            <NewState value="0"/>',
    '            <MarkIndex value="65535"/>',
    '            <CurrentIndex value="65535"/>',
    "          </Transition>",
    '          <Transition onGlyphClass="3">',
    '            <NewState value="0"/>',
    '            <MarkIndex value="65535"/>',
    '            <CurrentIndex value="65535"/>',
    "          </Transition>",
    '          <Transition onGlyphClass="4">',
    '            <NewState value="0"/>',
    '            <MarkIndex value="65535"/>',
    '            <CurrentIndex value="65535"/>',
    "          </Transition>",
    '          <Transition onGlyphClass="5">',
    '            <NewState value="2"/>',
    '            <MarkIndex value="65535"/>',
    '            <CurrentIndex value="65535"/>',
    "          </Transition>",
    "        </State>",
    '        <State index="1">',
    '          <Transition onGlyphClass="0">',
    '            <NewState value="0"/>',
    '            <MarkIndex value="65535"/>',
    '            <CurrentIndex value="65535"/>',
    "          </Transition>",
    '          <Transition onGlyphClass="1">',
    '            <NewState value="0"/>',
    '            <MarkIndex value="65535"/>',
    '            <CurrentIndex value="65535"/>',
    "          </Transition>",
    '          <Transition onGlyphClass="2">',
    '            <NewState value="0"/>',
    '            <MarkIndex value="65535"/>',
    '            <CurrentIndex value="65535"/>',
    "          </Transition>",
    '          <Transition onGlyphClass="3">',
    '            <NewState value="0"/>',
    '            <MarkIndex value="65535"/>',
    '            <CurrentIndex value="65535"/>',
    "          </Transition>",
    '          <Transition onGlyphClass="4">',
    '            <NewState value="0"/>',
    '            <MarkIndex value="65535"/>',
    '            <CurrentIndex value="65535"/>',
    "          </Transition>",
    '          <Transition onGlyphClass="5">',
    '            <NewState value="2"/>',
    '            <MarkIndex value="65535"/>',
    '            <CurrentIndex value="65535"/>',
    "          </Transition>",
    "        </State>",
    '        <State index="2">',
    '          <Transition onGlyphClass="0">',
    '            <NewState value="0"/>',
    '            <MarkIndex value="65535"/>',
    '            <CurrentIndex value="65535"/>',
    "          </Transition>",
    '          <Transition onGlyphClass="1">',
    '            <NewState value="0"/>',
    '            <MarkIndex value="65535"/>',
    '            <CurrentIndex value="65535"/>',
    "          </Transition>",
    '          <Transition onGlyphClass="2">',
    '            <NewState value="0"/>',
    '            <MarkIndex value="65535"/>',
    '            <CurrentIndex value="65535"/>',
    "          </Transition>",
    '          <Transition onGlyphClass="3">',
    '            <NewState value="0"/>',
    '            <MarkIndex value="65535"/>',
    '            <CurrentIndex value="65535"/>',
    "          </Transition>",
    '          <Transition onGlyphClass="4">',
    '            <NewState value="0"/>',
    '            <MarkIndex value="65535"/>',
    '            <CurrentIndex value="0"/>',
    "          </Transition>",
    '          <Transition onGlyphClass="5">',
    '            <NewState value="2"/>',
    '            <MarkIndex value="65535"/>',
    '            <CurrentIndex value="65535"/>',
    "          </Transition>",
    "        </State>",
    '        <PerGlyphLookup index="0">',
    '          <Lookup glyph="A" value="A.swash"/>',
    '          <Lookup glyph="B" value="B.swash"/>',
    '          <Lookup glyph="X" value="X.swash"/>',
    '          <Lookup glyph="Y" value="Y.swash"/>',
    "        </PerGlyphLookup>",
    "      </StateTable>",
    "    </ContextualMorph>",
    "  </MorphSubtable>",
    "</MorphChain>",
]


# Taken from “Example 2: A ligature table” in
# https://developer.apple.com/fonts/TrueType-Reference-Manual/RM06/Chap6morx.html
# as retrieved on 2017-09-11.
#
# Compared to the example table in Apple’s specification, we’ve
# made the following changes:
#
# * at offsets 0..35, we’ve prepended 36 bytes of boilerplate
#   to make the data a structurally valid ‘morx’ table;
#
# * at offsets 88..91 (offsets 52..55 in Apple’s document), we’ve
#   changed the range of the third segment from 23..24 to 26..28.
#   The hexdump values in Apple’s specification are completely wrong;
#   the values from the comments would work, but they can be encoded
#   more compactly than in the specification example. For round-trip
#   testing, we omit the ‘f’ glyph, which makes AAT lookup format 2
#   the most compact encoding;
#
# * at offsets 92..93 (offsets 56..57 in Apple’s document), we’ve
#   changed the glyph class of the third segment from 5 to 6, which
#   matches the values from the comments to the spec (but not the
#   Apple’s hexdump).
#
# TODO: Ask Apple to fix “Example 2” in the ‘morx’ specification.
MORX_LIGATURE_DATA = deHexStr(
    "0002 0000 "  #  0: Version=2, Reserved=0
    "0000 0001 "  #  4: MorphChainCount=1
    "0000 0001 "  #  8: DefaultFlags=1
    "0000 00DA "  # 12: StructLength=218 (+8=226)
    "0000 0000 "  # 16: MorphFeatureCount=0
    "0000 0001 "  # 20: MorphSubtableCount=1
    "0000 00CA "  # 24: Subtable[0].StructLength=202 (+24=226)
    "80 "  # 28: Subtable[0].CoverageFlags=0x80
    "00 00 "  # 29: Subtable[0].Reserved=0
    "02 "  # 31: Subtable[0].MorphType=2/LigatureMorph
    "0000 0001 "  # 32: Subtable[0].SubFeatureFlags=0x1
    # State table header.
    "0000 0007 "  # 36: STXHeader.ClassCount=7
    "0000 001C "  # 40: STXHeader.ClassTableOffset=28 (+36=64)
    "0000 0040 "  # 44: STXHeader.StateArrayOffset=64 (+36=100)
    "0000 0078 "  # 48: STXHeader.EntryTableOffset=120 (+36=156)
    "0000 0090 "  # 52: STXHeader.LigActionsOffset=144 (+36=180)
    "0000 009C "  # 56: STXHeader.LigComponentsOffset=156 (+36=192)
    "0000 00AE "  # 60: STXHeader.LigListOffset=174 (+36=210)
    # Glyph class table.
    "0002 0006 "  # 64: ClassTable.LookupFormat=2, .UnitSize=6
    "0003 000C "  # 68:   .NUnits=3, .SearchRange=12
    "0001 0006 "  # 72:   .EntrySelector=1, .RangeShift=6
    "0016 0014 0004 "  # 76: GlyphID 20..22 [a..c] -> GlyphClass 4
    "0018 0017 0005 "  # 82: GlyphID 23..24 [d..e] -> GlyphClass 5
    "001C 001A 0006 "  # 88: GlyphID 26..28 [g..i] -> GlyphClass 6
    "FFFF FFFF 0000 "  # 94: <end of lookup>
    # State array.
    "0000 0000 0000 0000 0001 0000 0000 "  # 100: State[0][0..6]
    "0000 0000 0000 0000 0001 0000 0000 "  # 114: State[1][0..6]
    "0000 0000 0000 0000 0001 0002 0000 "  # 128: State[2][0..6]
    "0000 0000 0000 0000 0001 0002 0003 "  # 142: State[3][0..6]
    # Entry table.
    "0000 0000 "  # 156: Entries[0].NewState=0, .Flags=0
    "0000 "  # 160: Entries[0].ActionIndex=<n/a> because no 0x2000 flag
    "0002 8000 "  # 162: Entries[1].NewState=2, .Flags=0x8000 (SetComponent)
    "0000 "  # 166: Entries[1].ActionIndex=<n/a> because no 0x2000 flag
    "0003 8000 "  # 168: Entries[2].NewState=3, .Flags=0x8000 (SetComponent)
    "0000 "  # 172: Entries[2].ActionIndex=<n/a> because no 0x2000 flag
    "0000 A000 "  # 174: Entries[3].NewState=0, .Flags=0xA000 (SetComponent,Act)
    "0000 "  # 178: Entries[3].ActionIndex=0 (start at Action[0])
    # Ligature actions table.
    "3FFF FFE7 "  # 180: Action[0].Flags=0, .GlyphIndexDelta=-25
    "3FFF FFED "  # 184: Action[1].Flags=0, .GlyphIndexDelta=-19
    "BFFF FFF2 "  # 188: Action[2].Flags=<end of list>, .GlyphIndexDelta=-14
    # Ligature component table.
    "0000 0001 "  # 192: LigComponent[0]=0, LigComponent[1]=1
    "0002 0003 "  # 196: LigComponent[2]=2, LigComponent[3]=3
    "0000 0004 "  # 200: LigComponent[4]=0, LigComponent[5]=4
    "0000 0008 "  # 204: LigComponent[6]=0, LigComponent[7]=8
    "0010      "  # 208: LigComponent[8]=16
    # Ligature list.
    "03E8 03E9 "  # 210: LigList[0]=1000, LigList[1]=1001
    "03EA 03EB "  # 214: LigList[2]=1002, LigList[3]=1003
    "03EC 03ED "  # 218: LigList[4]=1004, LigList[3]=1005
    "03EE 03EF "  # 222: LigList[5]=1006, LigList[6]=1007
)  # 226: <end>
assert len(MORX_LIGATURE_DATA) == 226, len(MORX_LIGATURE_DATA)


MORX_LIGATURE_XML = [
    '<Version value="2"/>',
    '<Reserved value="0"/>',
    "<!-- MorphChainCount=1 -->",
    '<MorphChain index="0">',
    '  <DefaultFlags value="0x00000001"/>',
    "  <!-- StructLength=218 -->",
    "  <!-- MorphFeatureCount=0 -->",
    "  <!-- MorphSubtableCount=1 -->",
    '  <MorphSubtable index="0">',
    "    <!-- StructLength=202 -->",
    '    <TextDirection value="Vertical"/>',
    '    <ProcessingOrder value="LayoutOrder"/>',
    "    <!-- MorphType=2 -->",
    '    <SubFeatureFlags value="0x00000001"/>',
    "    <LigatureMorph>",
    "      <StateTable>",
    "        <!-- GlyphClassCount=7 -->",
    '        <GlyphClass glyph="a" value="4"/>',
    '        <GlyphClass glyph="b" value="4"/>',
    '        <GlyphClass glyph="c" value="4"/>',
    '        <GlyphClass glyph="d" value="5"/>',
    '        <GlyphClass glyph="e" value="5"/>',
    '        <GlyphClass glyph="g" value="6"/>',
    '        <GlyphClass glyph="h" value="6"/>',
    '        <GlyphClass glyph="i" value="6"/>',
    '        <State index="0">',
    '          <Transition onGlyphClass="0">',
    '            <NewState value="0"/>',
    "          </Transition>",
    '          <Transition onGlyphClass="1">',
    '            <NewState value="0"/>',
    "          </Transition>",
    '          <Transition onGlyphClass="2">',
    '            <NewState value="0"/>',
    "          </Transition>",
    '          <Transition onGlyphClass="3">',
    '            <NewState value="0"/>',
    "          </Transition>",
    '          <Transition onGlyphClass="4">',
    '            <NewState value="2"/>',
    '            <Flags value="SetComponent"/>',
    "          </Transition>",
    '          <Transition onGlyphClass="5">',
    '            <NewState value="0"/>',
    "          </Transition>",
    '          <Transition onGlyphClass="6">',
    '            <NewState value="0"/>',
    "          </Transition>",
    "        </State>",
    '        <State index="1">',
    '          <Transition onGlyphClass="0">',
    '            <NewState value="0"/>',
    "          </Transition>",
    '          <Transition onGlyphClass="1">',
    '            <NewState value="0"/>',
    "          </Transition>",
    '          <Transition onGlyphClass="2">',
    '            <NewState value="0"/>',
    "          </Transition>",
    '          <Transition onGlyphClass="3">',
    '            <NewState value="0"/>',
    "          </Transition>",
    '          <Transition onGlyphClass="4">',
    '            <NewState value="2"/>',
    '            <Flags value="SetComponent"/>',
    "          </Transition>",
    '          <Transition onGlyphClass="5">',
    '            <NewState value="0"/>',
    "          </Transition>",
    '          <Transition onGlyphClass="6">',
    '            <NewState value="0"/>',
    "          </Transition>",
    "        </State>",
    '        <State index="2">',
    '          <Transition onGlyphClass="0">',
    '            <NewState value="0"/>',
    "          </Transition>",
    '          <Transition onGlyphClass="1">',
    '            <NewState value="0"/>',
    "          </Transition>",
    '          <Transition onGlyphClass="2">',
    '            <NewState value="0"/>',
    "          </Transition>",
    '          <Transition onGlyphClass="3">',
    '            <NewState value="0"/>',
    "          </Transition>",
    '          <Transition onGlyphClass="4">',
    '            <NewState value="2"/>',
    '            <Flags value="SetComponent"/>',
    "          </Transition>",
    '          <Transition onGlyphClass="5">',
    '            <NewState value="3"/>',
    '            <Flags value="SetComponent"/>',
    "          </Transition>",
    '          <Transition onGlyphClass="6">',
    '            <NewState value="0"/>',
    "          </Transition>",
    "        </State>",
    '        <State index="3">',
    '          <Transition onGlyphClass="0">',
    '            <NewState value="0"/>',
    "          </Transition>",
    '          <Transition onGlyphClass="1">',
    '            <NewState value="0"/>',
    "          </Transition>",
    '          <Transition onGlyphClass="2">',
    '            <NewState value="0"/>',
    "          </Transition>",
    '          <Transition onGlyphClass="3">',
    '            <NewState value="0"/>',
    "          </Transition>",
    '          <Transition onGlyphClass="4">',
    '            <NewState value="2"/>',
    '            <Flags value="SetComponent"/>',
    "          </Transition>",
    '          <Transition onGlyphClass="5">',
    '            <NewState value="3"/>',
    '            <Flags value="SetComponent"/>',
    "          </Transition>",
    '          <Transition onGlyphClass="6">',
    '            <NewState value="0"/>',
    '            <Flags value="SetComponent"/>',
    '            <Action GlyphIndexDelta="-25"/>',
    '            <Action GlyphIndexDelta="-19"/>',
    '            <Action GlyphIndexDelta="-14"/>',
    "          </Transition>",
    "        </State>",
    "        <LigComponents>",
    '          <LigComponent index="0" value="0"/>',
    '          <LigComponent index="1" value="1"/>',
    '          <LigComponent index="2" value="2"/>',
    '          <LigComponent index="3" value="3"/>',
    '          <LigComponent index="4" value="0"/>',
    '          <LigComponent index="5" value="4"/>',
    '          <LigComponent index="6" value="0"/>',
    '          <LigComponent index="7" value="8"/>',
    '          <LigComponent index="8" value="16"/>',
    "        </LigComponents>",
    "        <Ligatures>",
    '          <Ligature glyph="adf" index="0"/>',
    '          <Ligature glyph="adg" index="1"/>',
    '          <Ligature glyph="adh" index="2"/>',
    '          <Ligature glyph="adi" index="3"/>',
    '          <Ligature glyph="aef" index="4"/>',
    '          <Ligature glyph="aeg" index="5"/>',
    '          <Ligature glyph="aeh" index="6"/>',
    '          <Ligature glyph="aei" index="7"/>',
    "        </Ligatures>",
    "      </StateTable>",
    "    </LigatureMorph>",
    "  </MorphSubtable>",
    "</MorphChain>",
]


# Taken from the `morx` table of the second font in DevanagariSangamMN.ttc
# on macOS X 10.12.6; manually pruned to just contain the insertion lookup.
MORX_INSERTION_DATA = deHexStr(
    "0002 0000 "  #  0: Version=2, Reserved=0
    "0000 0001 "  #  4: MorphChainCount=1
    "0000 0001 "  #  8: DefaultFlags=1
    "0000 00A4 "  # 12: StructLength=164 (+8=172)
    "0000 0000 "  # 16: MorphFeatureCount=0
    "0000 0001 "  # 20: MorphSubtableCount=1
    "0000 0094 "  # 24: Subtable[0].StructLength=148 (+24=172)
    "00 "  # 28: Subtable[0].CoverageFlags=0x00
    "00 00 "  # 29: Subtable[0].Reserved=0
    "05 "  # 31: Subtable[0].MorphType=5/InsertionMorph
    "0000 0001 "  # 32: Subtable[0].SubFeatureFlags=0x1
    "0000 0006 "  # 36: STXHeader.ClassCount=6
    "0000 0014 "  # 40: STXHeader.ClassTableOffset=20 (+36=56)
    "0000 004A "  # 44: STXHeader.StateArrayOffset=74 (+36=110)
    "0000 006E "  # 48: STXHeader.EntryTableOffset=110 (+36=146)
    "0000 0086 "  # 52: STXHeader.InsertionActionOffset=134 (+36=170)
    # Glyph class table.
    "0002 0006 "  #  56: ClassTable.LookupFormat=2, .UnitSize=6
    "0006 0018 "  #  60:   .NUnits=6, .SearchRange=24
    "0002 000C "  #  64:   .EntrySelector=2, .RangeShift=12
    "00AC 00AC 0005 "  #  68: GlyphID 172..172 -> GlyphClass 5
    "01EB 01E6 0005 "  #  74: GlyphID 486..491 -> GlyphClass 5
    "01F0 01F0 0004 "  #  80: GlyphID 496..496 -> GlyphClass 4
    "01F8 01F6 0004 "  #  88: GlyphID 502..504 -> GlyphClass 4
    "01FC 01FA 0004 "  #  92: GlyphID 506..508 -> GlyphClass 4
    "0250 0250 0005 "  #  98: GlyphID 592..592 -> GlyphClass 5
    "FFFF FFFF 0000 "  # 104: <end of lookup>
    # State array.
    "0000 0000 0000 0000 0001 0000 "  # 110: State[0][0..5]
    "0000 0000 0000 0000 0001 0000 "  # 122: State[1][0..5]
    "0000 0000 0001 0000 0001 0002 "  # 134: State[2][0..5]
    # Entry table.
    "0000 0000 "  # 146: Entries[0].NewState=0, .Flags=0
    "FFFF "  # 150: Entries[0].CurrentInsertIndex=<None>
    "FFFF "  # 152: Entries[0].MarkedInsertIndex=<None>
    "0002 0000 "  # 154: Entries[1].NewState=0, .Flags=0
    "FFFF "  # 158: Entries[1].CurrentInsertIndex=<None>
    "FFFF "  # 160: Entries[1].MarkedInsertIndex=<None>
    "0000 "  # 162: Entries[2].NewState=0
    "2820 "  # 164:   .Flags=CurrentIsKashidaLike,CurrentInsertBefore
    #        .CurrentInsertCount=1, .MarkedInsertCount=0
    "0000 "  # 166: Entries[1].CurrentInsertIndex=0
    "FFFF "  # 168: Entries[1].MarkedInsertIndex=<None>
    # Insertion action table.
    "022F"  # 170: InsertionActionTable[0]=GlyphID 559
)  # 172: <end>
assert len(MORX_INSERTION_DATA) == 172, len(MORX_INSERTION_DATA)


MORX_INSERTION_XML = [
    '<Version value="2"/>',
    '<Reserved value="0"/>',
    "<!-- MorphChainCount=1 -->",
    '<MorphChain index="0">',
    '  <DefaultFlags value="0x00000001"/>',
    "  <!-- StructLength=164 -->",
    "  <!-- MorphFeatureCount=0 -->",
    "  <!-- MorphSubtableCount=1 -->",
    '  <MorphSubtable index="0">',
    "    <!-- StructLength=148 -->",
    '    <TextDirection value="Horizontal"/>',
    '    <ProcessingOrder value="LayoutOrder"/>',
    "    <!-- MorphType=5 -->",
    '    <SubFeatureFlags value="0x00000001"/>',
    "    <InsertionMorph>",
    "      <StateTable>",
    "        <!-- GlyphClassCount=6 -->",
    '        <GlyphClass glyph="g.172" value="5"/>',
    '        <GlyphClass glyph="g.486" value="5"/>',
    '        <GlyphClass glyph="g.487" value="5"/>',
    '        <GlyphClass glyph="g.488" value="5"/>',
    '        <GlyphClass glyph="g.489" value="5"/>',
    '        <GlyphClass glyph="g.490" value="5"/>',
    '        <GlyphClass glyph="g.491" value="5"/>',
    '        <GlyphClass glyph="g.496" value="4"/>',
    '        <GlyphClass glyph="g.502" value="4"/>',
    '        <GlyphClass glyph="g.503" value="4"/>',
    '        <GlyphClass glyph="g.504" value="4"/>',
    '        <GlyphClass glyph="g.506" value="4"/>',
    '        <GlyphClass glyph="g.507" value="4"/>',
    '        <GlyphClass glyph="g.508" value="4"/>',
    '        <GlyphClass glyph="g.592" value="5"/>',
    '        <State index="0">',
    '          <Transition onGlyphClass="0">',
    '            <NewState value="0"/>',
    "          </Transition>",
    '          <Transition onGlyphClass="1">',
    '            <NewState value="0"/>',
    "          </Transition>",
    '          <Transition onGlyphClass="2">',
    '            <NewState value="0"/>',
    "          </Transition>",
    '          <Transition onGlyphClass="3">',
    '            <NewState value="0"/>',
    "          </Transition>",
    '          <Transition onGlyphClass="4">',
    '            <NewState value="2"/>',
    "          </Transition>",
    '          <Transition onGlyphClass="5">',
    '            <NewState value="0"/>',
    "          </Transition>",
    "        </State>",
    '        <State index="1">',
    '          <Transition onGlyphClass="0">',
    '            <NewState value="0"/>',
    "          </Transition>",
    '          <Transition onGlyphClass="1">',
    '            <NewState value="0"/>',
    "          </Transition>",
    '          <Transition onGlyphClass="2">',
    '            <NewState value="0"/>',
    "          </Transition>",
    '          <Transition onGlyphClass="3">',
    '            <NewState value="0"/>',
    "          </Transition>",
    '          <Transition onGlyphClass="4">',
    '            <NewState value="2"/>',
    "          </Transition>",
    '          <Transition onGlyphClass="5">',
    '            <NewState value="0"/>',
    "          </Transition>",
    "        </State>",
    '        <State index="2">',
    '          <Transition onGlyphClass="0">',
    '            <NewState value="0"/>',
    "          </Transition>",
    '          <Transition onGlyphClass="1">',
    '            <NewState value="0"/>',
    "          </Transition>",
    '          <Transition onGlyphClass="2">',
    '            <NewState value="2"/>',
    "          </Transition>",
    '          <Transition onGlyphClass="3">',
    '            <NewState value="0"/>',
    "          </Transition>",
    '          <Transition onGlyphClass="4">',
    '            <NewState value="2"/>',
    "          </Transition>",
    '          <Transition onGlyphClass="5">',
    '            <NewState value="0"/>',
    '            <Flags value="CurrentIsKashidaLike,CurrentInsertBefore"/>',
    '            <CurrentInsertionAction glyph="g.559"/>',
    "          </Transition>",
    "        </State>",
    "      </StateTable>",
    "    </InsertionMorph>",
    "  </MorphSubtable>",
    "</MorphChain>",
]


class MORXNoncontextualGlyphSubstitutionTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.maxDiff = None
        glyphs = [".notdef"] + ["g.%d" % i for i in range(1, 140)]
        glyphs[11], glyphs[13] = "parenleft", "parenright"
        glyphs[135], glyphs[136] = "parenleft.vertical", "parenright.vertical"
        cls.font = FakeFont(glyphs)

    def test_decompile_toXML(self):
        table = newTable("morx")
        table.decompile(MORX_NONCONTEXTUAL_DATA, self.font)
        self.assertEqual(getXML(table.toXML), MORX_NONCONTEXTUAL_XML)

    def test_compile_fromXML(self):
        table = newTable("morx")
        for name, attrs, content in parseXML(MORX_NONCONTEXTUAL_XML):
            table.fromXML(name, attrs, content, font=self.font)
        self.assertEqual(
            hexStr(table.compile(self.font)), hexStr(MORX_NONCONTEXTUAL_DATA)
        )


class MORXRearrangementTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.maxDiff = None
        cls.font = FakeFont([".nodef", "A", "B", "C"])

    def test_decompile_toXML(self):
        table = newTable("morx")
        table.decompile(MORX_REARRANGEMENT_DATA, self.font)
        self.assertEqual(getXML(table.toXML), MORX_REARRANGEMENT_XML)

    def test_compile_fromXML(self):
        table = newTable("morx")
        for name, attrs, content in parseXML(MORX_REARRANGEMENT_XML):
            table.fromXML(name, attrs, content, font=self.font)
        self.assertEqual(
            hexStr(table.compile(self.font)), hexStr(MORX_REARRANGEMENT_DATA)
        )


class MORXContextualSubstitutionTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.maxDiff = None
        g = [".notdef"] + ["g.%d" % i for i in range(1, 910)]
        g[80] = "C"
        g[50], g[52], g[201], g[202] = "A", "B", "X", "Y"
        g[600], g[601], g[602], g[900] = ("A.swash", "B.swash", "X.swash", "Y.swash")
        cls.font = FakeFont(g)

    def test_decompile_toXML(self):
        table = newTable("morx")
        table.decompile(MORX_CONTEXTUAL_DATA, self.font)
        self.assertEqual(getXML(table.toXML), MORX_CONTEXTUAL_XML)

    def test_compile_fromXML(self):
        table = newTable("morx")
        for name, attrs, content in parseXML(MORX_CONTEXTUAL_XML):
            table.fromXML(name, attrs, content, font=self.font)
        self.assertEqual(hexStr(table.compile(self.font)), hexStr(MORX_CONTEXTUAL_DATA))


class MORXLigatureSubstitutionTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.maxDiff = None
        g = [".notdef"] + ["g.%d" % i for i in range(1, 1515)]
        g[20:29] = "a b c d e f g h i".split()
        g[1000:1008] = "adf adg adh adi aef aeg aeh aei".split()
        g[1008:1016] = "bdf bdg bdh bdi bef beg beh bei".split()
        g[1500:1507] = "cdf cdg cdh cdi cef ceg ceh".split()
        g[1511] = "cei"
        cls.font = FakeFont(g)

    def test_decompile_toXML(self):
        table = newTable("morx")
        table.decompile(MORX_LIGATURE_DATA, self.font)
        self.assertEqual(getXML(table.toXML), MORX_LIGATURE_XML)

    def test_compile_fromXML(self):
        table = newTable("morx")
        for name, attrs, content in parseXML(MORX_LIGATURE_XML):
            table.fromXML(name, attrs, content, font=self.font)
        self.assertEqual(hexStr(table.compile(self.font)), hexStr(MORX_LIGATURE_DATA))


class MORXGlyphInsertionTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.maxDiff = None
        cls.font = FakeFont([".notdef"] + ["g.%d" % i for i in range(1, 910)])

    def test_decompile_toXML(self):
        table = newTable("morx")
        table.decompile(MORX_INSERTION_DATA, self.font)
        self.assertEqual(getXML(table.toXML), MORX_INSERTION_XML)

    def test_compile_fromXML(self):
        table = newTable("morx")
        for name, attrs, content in parseXML(MORX_INSERTION_XML):
            table.fromXML(name, attrs, content, font=self.font)
        self.assertEqual(hexStr(table.compile(self.font)), hexStr(MORX_INSERTION_DATA))


class MORXCoverageFlagsTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.maxDiff = None
        cls.font = FakeFont([".notdef", "A", "B", "C"])

    def checkFlags(self, flags, textDirection, processingOrder, checkCompile=True):
        data = bytesjoin(
            [
                MORX_REARRANGEMENT_DATA[:28],
                bytechr(flags << 4),
                MORX_REARRANGEMENT_DATA[29:],
            ]
        )
        xml = []
        for line in MORX_REARRANGEMENT_XML:
            if line.startswith("    <TextDirection "):
                line = '    <TextDirection value="%s"/>' % textDirection
            elif line.startswith("    <ProcessingOrder "):
                line = '    <ProcessingOrder value="%s"/>' % processingOrder
            xml.append(line)
        table1 = newTable("morx")
        table1.decompile(data, self.font)
        self.assertEqual(getXML(table1.toXML), xml)
        if checkCompile:
            table2 = newTable("morx")
            for name, attrs, content in parseXML(xml):
                table2.fromXML(name, attrs, content, font=self.font)
            self.assertEqual(hexStr(table2.compile(self.font)), hexStr(data))

    def test_CoverageFlags(self):
        self.checkFlags(0x0, "Horizontal", "LayoutOrder")
        self.checkFlags(0x1, "Horizontal", "LogicalOrder")
        self.checkFlags(0x2, "Any", "LayoutOrder")
        self.checkFlags(0x3, "Any", "LogicalOrder")
        self.checkFlags(0x4, "Horizontal", "ReversedLayoutOrder")
        self.checkFlags(0x5, "Horizontal", "ReversedLogicalOrder")
        self.checkFlags(0x6, "Any", "ReversedLayoutOrder")
        self.checkFlags(0x7, "Any", "ReversedLogicalOrder")
        self.checkFlags(0x8, "Vertical", "LayoutOrder")
        self.checkFlags(0x9, "Vertical", "LogicalOrder")
        # We do not always check the compilation to binary data:
        # some flag combinations do not make sense to emit in binary.
        # Specifically, if bit 28 (TextDirection=Any) is set in
        # CoverageFlags, bit 30 (TextDirection=Vertical) is to be
        # ignored according to the 'morx' specification. We still want
        # to test the _decoding_ of 'morx' subtables whose CoverageFlags
        # have both bits 28 and 30 set, since this is a valid flag
        # combination with defined semantics.  However, our encoder
        # does not set TextDirection=Vertical when TextDirection=Any.
        self.checkFlags(0xA, "Any", "LayoutOrder", checkCompile=False)
        self.checkFlags(0xB, "Any", "LogicalOrder", checkCompile=False)
        self.checkFlags(0xC, "Vertical", "ReversedLayoutOrder")
        self.checkFlags(0xD, "Vertical", "ReversedLogicalOrder")
        self.checkFlags(0xE, "Any", "ReversedLayoutOrder", checkCompile=False)
        self.checkFlags(0xF, "Any", "ReversedLogicalOrder", checkCompile=False)

    def test_ReservedCoverageFlags(self):
        # 8A BC DE = TextDirection=Vertical, Reserved=0xABCDE
        # Note that the lower 4 bits of the first byte are already
        # part of the Reserved value. We test the full round-trip
        # to encoding and decoding is quite hairy.
        data = bytesjoin(
            [
                MORX_REARRANGEMENT_DATA[:28],
                bytechr(0x8A),
                bytechr(0xBC),
                bytechr(0xDE),
                MORX_REARRANGEMENT_DATA[31:],
            ]
        )
        table = newTable("morx")
        table.decompile(data, self.font)
        subtable = table.table.MorphChain[0].MorphSubtable[0]
        self.assertEqual(subtable.Reserved, 0xABCDE)
        xml = getXML(table.toXML)
        self.assertIn('    <Reserved value="0xabcde"/>', xml)
        table2 = newTable("morx")
        for name, attrs, content in parseXML(xml):
            table2.fromXML(name, attrs, content, font=self.font)
        self.assertEqual(hexStr(table2.compile(self.font)[28:31]), "8abcde")


class UnsupportedMorxLookupTest(unittest.TestCase):
    def __init__(self, methodName):
        unittest.TestCase.__init__(self, methodName)
        # Python 3 renamed assertRaisesRegexp to assertRaisesRegex,
        # and fires deprecation warnings if a program uses the old name.
        if not hasattr(self, "assertRaisesRegex"):
            self.assertRaisesRegex = self.assertRaisesRegexp

    def test_unsupportedLookupType(self):
        data = bytesjoin(
            [MORX_NONCONTEXTUAL_DATA[:67], bytechr(66), MORX_NONCONTEXTUAL_DATA[69:]]
        )
        with self.assertRaisesRegex(
            AssertionError, r"unsupported 'morx' lookup type 66"
        ):
            morx = newTable("morx")
            morx.decompile(data, FakeFont([".notdef"]))


if __name__ == "__main__":
    import sys

    sys.exit(unittest.main())



from fontTools.pens.transformPen import TransformPen
from fontTools.misc import etree
from fontTools.misc.textTools import tostr
from .parser import parse_path
from .shapes import PathBuilder


__all__ = [tostr(s) for s in ("SVGPath", "parse_path")]


class SVGPath(object):
    """Parse SVG ``path`` elements from a file or string, and draw them
    onto a glyph object that supports the FontTools Pen protocol.

    For example, reading from an SVG file and drawing to a Defcon Glyph:

    .. code-block::

        import defcon
        glyph = defcon.Glyph()
        pen = glyph.getPen()
        svg = SVGPath("path/to/a.svg")
        svg.draw(pen)

    Or reading from a string containing SVG data, using the alternative
    'fromstring' (a class method):

    .. code-block::

        data = '<?xml version="1.0" ...'
        svg = SVGPath.fromstring(data)
        svg.draw(pen)

    Both constructors can optionally take a 'transform' matrix (6-float
    tuple, or a FontTools Transform object) to modify the draw output.
    """

    def __init__(self, filename=None, transform=None):
        if filename is None:
            self.root = etree.ElementTree()
        else:
            tree = etree.parse(filename)
            self.root = tree.getroot()
        self.transform = transform

    @classmethod
    def fromstring(cls, data, transform=None):
        self = cls(transform=transform)
        self.root = etree.fromstring(data)
        return self

    def draw(self, pen):
        if self.transform:
            pen = TransformPen(pen, self.transform)
        pb = PathBuilder()
        # xpath | doesn't seem to reliable work so just walk it
        for el in self.root.iter():
            pb.add_path_from_element(el)
        original_pen = pen
        for path, transform in zip(pb.paths, pb.transforms):
            if transform:
                pen = TransformPen(original_pen, transform)
            else:
                pen = original_pen
            parse_path(path, pen)



from __future__ import annotations

from typing import TYPE_CHECKING

from fontTools.misc.textTools import Tag
from fontTools.ttLib import getClassTag

if TYPE_CHECKING:
    from typing import Any

    from fontTools.misc.xmlWriter import XMLWriter
    from fontTools.ttLib import TTFont


class DefaultTable:
    dependencies: list[str] = []

    def __init__(self, tag: str | bytes | None = None) -> None:
        if tag is None:
            tag = getClassTag(self.__class__)
        self.tableTag = Tag(tag)

    def decompile(self, data: bytes, ttFont: TTFont) -> None:
        self.data = data

    def compile(self, ttFont: TTFont) -> bytes:
        return self.data

    def toXML(
        self, writer: XMLWriter, ttFont: TTFont, **kwargs: dict[str, Any]
    ) -> None:
        if hasattr(self, "ERROR"):
            writer.comment("An error occurred during the decompilation of this table")
            writer.newline()
            writer.comment(self.ERROR)
            writer.newline()
        writer.begintag("hexdata")
        writer.newline()
        writer.dumphex(self.compile(ttFont))
        writer.endtag("hexdata")
        writer.newline()

    def fromXML(
        self, name: str, attrs: dict[str, str], content: str, ttFont: TTFont
    ) -> None:
        from fontTools import ttLib
        from fontTools.misc.textTools import readHex

        if name != "hexdata":
            raise ttLib.TTLibError("can't handle '%s' element" % name)
        self.decompile(readHex(content), ttFont)

    def __repr__(self) -> str:
        return "<'%s' table at %x>" % (self.tableTag, id(self))

    def __eq__(self, other: Any) -> bool:
        if type(self) != type(other):
            return NotImplemented
        return self.__dict__ == other.__dict__

    def __ne__(self, other: Any) -> bool:
        result = self.__eq__(other)
        return result if result is NotImplemented else not result



from fontTools.misc.testTools import parseXML, getXML
from fontTools.misc.textTools import deHexStr
from fontTools.ttLib import TTFont, TTLibError
from fontTools.ttLib.tables._t_r_a_k import *
from fontTools.ttLib.tables._n_a_m_e import table__n_a_m_e, NameRecord
import unittest


# /Library/Fonts/Osaka.ttf from OSX has trak table with both horiz and vertData
OSAKA_TRAK_TABLE_DATA = deHexStr(
    "00 01 00 00 00 00 00 0c 00 40 00 00 00 03 00 02 00 00 00 2c ff ff "
    "00 00 01 06 00 34 00 00 00 00 01 07 00 38 00 01 00 00 01 08 00 3c "
    "00 0c 00 00 00 18 00 00 ff f4 ff f4 00 00 00 00 00 0c 00 0c 00 03 "
    "00 02 00 00 00 60 ff ff 00 00 01 09 00 68 00 00 00 00 01 0a 00 6c "
    "00 01 00 00 01 0b 00 70 00 0c 00 00 00 18 00 00 ff f4 ff f4 00 00 "
    "00 00 00 0c 00 0c"
)

# decompiled horizData and vertData entries from Osaka.ttf
OSAKA_HORIZ_TRACK_ENTRIES = {
    -1.0: TrackTableEntry({24.0: -12, 12.0: -12}, nameIndex=262),
    0.0: TrackTableEntry({24.0: 0, 12.0: 0}, nameIndex=263),
    1.0: TrackTableEntry({24.0: 12, 12.0: 12}, nameIndex=264),
}

OSAKA_VERT_TRACK_ENTRIES = {
    -1.0: TrackTableEntry({24.0: -12, 12.0: -12}, nameIndex=265),
    0.0: TrackTableEntry({24.0: 0, 12.0: 0}, nameIndex=266),
    1.0: TrackTableEntry({24.0: 12, 12.0: 12}, nameIndex=267),
}

OSAKA_TRAK_TABLE_XML = [
    '<version value="1.0"/>',
    '<format value="0"/>',
    "<horizData>",
    "  <!-- nTracks=3, nSizes=2 -->",
    '  <trackEntry value="-1.0" nameIndex="262">',
    "    <!-- Tight -->",
    '    <track size="12.0" value="-12"/>',
    '    <track size="24.0" value="-12"/>',
    "  </trackEntry>",
    '  <trackEntry value="0.0" nameIndex="263">',
    "    <!-- Normal -->",
    '    <track size="12.0" value="0"/>',
    '    <track size="24.0" value="0"/>',
    "  </trackEntry>",
    '  <trackEntry value="1.0" nameIndex="264">',
    "    <!-- Loose -->",
    '    <track size="12.0" value="12"/>',
    '    <track size="24.0" value="12"/>',
    "  </trackEntry>",
    "</horizData>",
    "<vertData>",
    "  <!-- nTracks=3, nSizes=2 -->",
    '  <trackEntry value="-1.0" nameIndex="265">',
    "    <!-- Tight -->",
    '    <track size="12.0" value="-12"/>',
    '    <track size="24.0" value="-12"/>',
    "  </trackEntry>",
    '  <trackEntry value="0.0" nameIndex="266">',
    "    <!-- Normal -->",
    '    <track size="12.0" value="0"/>',
    '    <track size="24.0" value="0"/>',
    "  </trackEntry>",
    '  <trackEntry value="1.0" nameIndex="267">',
    "    <!-- Loose -->",
    '    <track size="12.0" value="12"/>',
    '    <track size="24.0" value="12"/>',
    "  </trackEntry>",
    "</vertData>",
]

# made-up table containing only vertData (no horizData)
OSAKA_VERT_ONLY_TRAK_TABLE_DATA = deHexStr(
    "00 01 00 00 00 00 00 00 00 0c 00 00 00 03 00 02 00 00 00 2c ff ff "
    "00 00 01 09 00 34 00 00 00 00 01 0a 00 38 00 01 00 00 01 0b 00 3c "
    "00 0c 00 00 00 18 00 00 ff f4 ff f4 00 00 00 00 00 0c 00 0c"
)

OSAKA_VERT_ONLY_TRAK_TABLE_XML = [
    '<version value="1.0"/>',
    '<format value="0"/>',
    "<horizData>",
    "  <!-- nTracks=0, nSizes=0 -->",
    "</horizData>",
    "<vertData>",
    "  <!-- nTracks=3, nSizes=2 -->",
    '  <trackEntry value="-1.0" nameIndex="265">',
    "    <!-- Tight -->",
    '    <track size="12.0" value="-12"/>',
    '    <track size="24.0" value="-12"/>',
    "  </trackEntry>",
    '  <trackEntry value="0.0" nameIndex="266">',
    "    <!-- Normal -->",
    '    <track size="12.0" value="0"/>',
    '    <track size="24.0" value="0"/>',
    "  </trackEntry>",
    '  <trackEntry value="1.0" nameIndex="267">',
    "    <!-- Loose -->",
    '    <track size="12.0" value="12"/>',
    '    <track size="24.0" value="12"/>',
    "  </trackEntry>",
    "</vertData>",
]


# also /Library/Fonts/Skia.ttf contains a trak table with horizData
SKIA_TRAK_TABLE_DATA = deHexStr(
    "00 01 00 00 00 00 00 0c 00 00 00 00 00 03 00 05 00 00 00 2c ff ff "
    "00 00 01 13 00 40 00 00 00 00 01 2f 00 4a 00 01 00 00 01 14 00 54 "
    "00 09 00 00 00 0a 00 00 00 0c 00 00 00 12 00 00 00 13 00 00 ff f6 "
    "ff e2 ff c4 ff c1 ff c1 00 0f 00 00 ff fb ff e7 ff e7 00 8c 00 82 "
    "00 7d 00 73 00 73"
)

SKIA_TRACK_ENTRIES = {
    -1.0: TrackTableEntry(
        {9.0: -10, 10.0: -30, 19.0: -63, 12.0: -60, 18.0: -63}, nameIndex=275
    ),
    0.0: TrackTableEntry(
        {9.0: 15, 10.0: 0, 19.0: -25, 12.0: -5, 18.0: -25}, nameIndex=303
    ),
    1.0: TrackTableEntry(
        {9.0: 140, 10.0: 130, 19.0: 115, 12.0: 125, 18.0: 115}, nameIndex=276
    ),
}

SKIA_TRAK_TABLE_XML = [
    '<version value="1.0"/>',
    '<format value="0"/>',
    "<horizData>",
    "  <!-- nTracks=3, nSizes=5 -->",
    '  <trackEntry value="-1.0" nameIndex="275">',
    "    <!-- Tight -->",
    '    <track size="9.0" value="-10"/>',
    '    <track size="10.0" value="-30"/>',
    '    <track size="12.0" value="-60"/>',
    '    <track size="18.0" value="-63"/>',
    '    <track size="19.0" value="-63"/>',
    "  </trackEntry>",
    '  <trackEntry value="0.0" nameIndex="303">',
    "    <!-- Normal -->",
    '    <track size="9.0" value="15"/>',
    '    <track size="10.0" value="0"/>',
    '    <track size="12.0" value="-5"/>',
    '    <track size="18.0" value="-25"/>',
    '    <track size="19.0" value="-25"/>',
    "  </trackEntry>",
    '  <trackEntry value="1.0" nameIndex="276">',
    "    <!-- Loose -->",
    '    <track size="9.0" value="140"/>',
    '    <track size="10.0" value="130"/>',
    '    <track size="12.0" value="125"/>',
    '    <track size="18.0" value="115"/>',
    '    <track size="19.0" value="115"/>',
    "  </trackEntry>",
    "</horizData>",
    "<vertData>",
    "  <!-- nTracks=0, nSizes=0 -->",
    "</vertData>",
]


class TrackingTableTest(unittest.TestCase):
    def __init__(self, methodName):
        unittest.TestCase.__init__(self, methodName)
        # Python 3 renamed assertRaisesRegexp to assertRaisesRegex,
        # and fires deprecation warnings if a program uses the old name.
        if not hasattr(self, "assertRaisesRegex"):
            self.assertRaisesRegex = self.assertRaisesRegexp

    def setUp(self):
        table = table__t_r_a_k()
        table.version = 1.0
        table.format = 0
        self.font = {"trak": table}

    def test_compile_horiz(self):
        table = self.font["trak"]
        table.horizData = TrackData(SKIA_TRACK_ENTRIES)
        trakData = table.compile(self.font)
        self.assertEqual(trakData, SKIA_TRAK_TABLE_DATA)

    def test_compile_vert(self):
        table = self.font["trak"]
        table.vertData = TrackData(OSAKA_VERT_TRACK_ENTRIES)
        trakData = table.compile(self.font)
        self.assertEqual(trakData, OSAKA_VERT_ONLY_TRAK_TABLE_DATA)

    def test_compile_horiz_and_vert(self):
        table = self.font["trak"]
        table.horizData = TrackData(OSAKA_HORIZ_TRACK_ENTRIES)
        table.vertData = TrackData(OSAKA_VERT_TRACK_ENTRIES)
        trakData = table.compile(self.font)
        self.assertEqual(trakData, OSAKA_TRAK_TABLE_DATA)

    def test_compile_longword_aligned(self):
        table = self.font["trak"]
        # without padding, this 'horizData' would end up 46 byte long
        table.horizData = TrackData(
            {0.0: TrackTableEntry(nameIndex=256, values={12.0: 0, 24.0: 0, 36.0: 0})}
        )
        table.vertData = TrackData(
            {0.0: TrackTableEntry(nameIndex=257, values={12.0: 0, 24.0: 0, 36.0: 0})}
        )
        trakData = table.compile(self.font)
        self.assertTrue(table.vertOffset % 4 == 0)

    def test_compile_sizes_mismatch(self):
        table = self.font["trak"]
        table.horizData = TrackData(
            {
                -1.0: TrackTableEntry(nameIndex=256, values={9.0: -10, 10.0: -30}),
                0.0: TrackTableEntry(nameIndex=257, values={8.0: 20, 12.0: 0}),
            }
        )
        with self.assertRaisesRegex(TTLibError, "entries must specify the same sizes"):
            table.compile(self.font)

    def test_decompile_horiz(self):
        table = self.font["trak"]
        table.decompile(SKIA_TRAK_TABLE_DATA, self.font)
        self.assertEqual(table.horizData, SKIA_TRACK_ENTRIES)
        self.assertEqual(table.vertData, TrackData())

    def test_decompile_vert(self):
        table = self.font["trak"]
        table.decompile(OSAKA_VERT_ONLY_TRAK_TABLE_DATA, self.font)
        self.assertEqual(table.horizData, TrackData())
        self.assertEqual(table.vertData, OSAKA_VERT_TRACK_ENTRIES)

    def test_decompile_horiz_and_vert(self):
        table = self.font["trak"]
        table.decompile(OSAKA_TRAK_TABLE_DATA, self.font)
        self.assertEqual(table.horizData, OSAKA_HORIZ_TRACK_ENTRIES)
        self.assertEqual(table.vertData, OSAKA_VERT_TRACK_ENTRIES)

    def test_roundtrip_decompile_compile(self):
        for trakData in (
            OSAKA_TRAK_TABLE_DATA,
            OSAKA_VERT_ONLY_TRAK_TABLE_DATA,
            SKIA_TRAK_TABLE_DATA,
        ):
            table = table__t_r_a_k()
            table.decompile(trakData, ttFont=None)
            newTrakData = table.compile(ttFont=None)
            self.assertEqual(trakData, newTrakData)

    def test_fromXML_horiz(self):
        table = self.font["trak"]
        for name, attrs, content in parseXML(SKIA_TRAK_TABLE_XML):
            table.fromXML(name, attrs, content, self.font)
        self.assertEqual(table.version, 1.0)
        self.assertEqual(table.format, 0)
        self.assertEqual(table.horizData, SKIA_TRACK_ENTRIES)
        self.assertEqual(table.vertData, TrackData())

    def test_fromXML_horiz_and_vert(self):
        table = self.font["trak"]
        for name, attrs, content in parseXML(OSAKA_TRAK_TABLE_XML):
            table.fromXML(name, attrs, content, self.font)
        self.assertEqual(table.version, 1.0)
        self.assertEqual(table.format, 0)
        self.assertEqual(table.horizData, OSAKA_HORIZ_TRACK_ENTRIES)
        self.assertEqual(table.vertData, OSAKA_VERT_TRACK_ENTRIES)

    def test_fromXML_vert(self):
        table = self.font["trak"]
        for name, attrs, content in parseXML(OSAKA_VERT_ONLY_TRAK_TABLE_XML):
            table.fromXML(name, attrs, content, self.font)
        self.assertEqual(table.version, 1.0)
        self.assertEqual(table.format, 0)
        self.assertEqual(table.horizData, TrackData())
        self.assertEqual(table.vertData, OSAKA_VERT_TRACK_ENTRIES)

    def test_toXML_horiz(self):
        table = self.font["trak"]
        table.horizData = TrackData(SKIA_TRACK_ENTRIES)
        add_name(self.font, "Tight", nameID=275)
        add_name(self.font, "Normal", nameID=303)
        add_name(self.font, "Loose", nameID=276)
        self.assertEqual(SKIA_TRAK_TABLE_XML, getXML(table.toXML, self.font))

    def test_toXML_horiz_and_vert(self):
        table = self.font["trak"]
        table.horizData = TrackData(OSAKA_HORIZ_TRACK_ENTRIES)
        table.vertData = TrackData(OSAKA_VERT_TRACK_ENTRIES)
        add_name(self.font, "Tight", nameID=262)
        add_name(self.font, "Normal", nameID=263)
        add_name(self.font, "Loose", nameID=264)
        add_name(self.font, "Tight", nameID=265)
        add_name(self.font, "Normal", nameID=266)
        add_name(self.font, "Loose", nameID=267)
        self.assertEqual(OSAKA_TRAK_TABLE_XML, getXML(table.toXML, self.font))

    def test_toXML_vert(self):
        table = self.font["trak"]
        table.vertData = TrackData(OSAKA_VERT_TRACK_ENTRIES)
        add_name(self.font, "Tight", nameID=265)
        add_name(self.font, "Normal", nameID=266)
        add_name(self.font, "Loose", nameID=267)
        self.assertEqual(OSAKA_VERT_ONLY_TRAK_TABLE_XML, getXML(table.toXML, self.font))

    def test_roundtrip_fromXML_toXML(self):
        font = {}
        add_name(font, "Tight", nameID=275)
        add_name(font, "Normal", nameID=303)
        add_name(font, "Loose", nameID=276)
        add_name(font, "Tight", nameID=262)
        add_name(font, "Normal", nameID=263)
        add_name(font, "Loose", nameID=264)
        add_name(font, "Tight", nameID=265)
        add_name(font, "Normal", nameID=266)
        add_name(font, "Loose", nameID=267)
        for input_xml in (
            SKIA_TRAK_TABLE_XML,
            OSAKA_TRAK_TABLE_XML,
            OSAKA_VERT_ONLY_TRAK_TABLE_XML,
        ):
            table = table__t_r_a_k()
            font["trak"] = table
            for name, attrs, content in parseXML(input_xml):
                table.fromXML(name, attrs, content, font)
            output_xml = getXML(table.toXML, font)
            self.assertEqual(input_xml, output_xml)


def add_name(font, string, nameID):
    nameTable = font.get("name")
    if nameTable is None:
        nameTable = font["name"] = table__n_a_m_e()
        nameTable.names = []
    namerec = NameRecord()
    namerec.nameID = nameID
    namerec.string = string.encode("mac_roman")
    namerec.platformID, namerec.platEncID, namerec.langID = (1, 0, 0)
    nameTable.names.append(namerec)


if __name__ == "__main__":
    import sys

    sys.exit(unittest.main())



# Copyright 2013 Google, Inc. All Rights Reserved.
#
# Google Author(s): Matt Fontaine


from fontTools.misc.textTools import bytesjoin
from fontTools.misc import sstruct
from . import E_B_D_T_
from .BitmapGlyphMetrics import (
    BigGlyphMetrics,
    bigGlyphMetricsFormat,
    SmallGlyphMetrics,
    smallGlyphMetricsFormat,
)
from .E_B_D_T_ import (
    BitmapGlyph,
    BitmapPlusSmallMetricsMixin,
    BitmapPlusBigMetricsMixin,
)
import struct


class table_C_B_D_T_(E_B_D_T_.table_E_B_D_T_):
    """Color Bitmap Data table

    The ``CBDT`` table contains color bitmap data for glyphs. It must
    be used in concert with the ``CBLC`` table.

    It is backwards-compatible with the monochrome/grayscale ``EBDT`` table.

    See also https://learn.microsoft.com/en-us/typography/opentype/spec/cbdt
    """

    # Change the data locator table being referenced.
    locatorName = "CBLC"

    # Modify the format class accessor for color bitmap use.
    def getImageFormatClass(self, imageFormat):
        try:
            return E_B_D_T_.table_E_B_D_T_.getImageFormatClass(self, imageFormat)
        except KeyError:
            return cbdt_bitmap_classes[imageFormat]


# Helper method for removing export features not supported by color bitmaps.
# Write data in the parent class will default to raw if an option is unsupported.
def _removeUnsupportedForColor(dataFunctions):
    dataFunctions = dict(dataFunctions)
    del dataFunctions["row"]
    return dataFunctions


class ColorBitmapGlyph(BitmapGlyph):
    fileExtension = ".png"
    xmlDataFunctions = _removeUnsupportedForColor(BitmapGlyph.xmlDataFunctions)


class cbdt_bitmap_format_17(BitmapPlusSmallMetricsMixin, ColorBitmapGlyph):
    def decompile(self):
        self.metrics = SmallGlyphMetrics()
        dummy, data = sstruct.unpack2(smallGlyphMetricsFormat, self.data, self.metrics)
        (dataLen,) = struct.unpack(">L", data[:4])
        data = data[4:]

        # For the image data cut it to the size specified by dataLen.
        assert dataLen <= len(data), "Data overun in format 17"
        self.imageData = data[:dataLen]

    def compile(self, ttFont):
        dataList = []
        dataList.append(sstruct.pack(smallGlyphMetricsFormat, self.metrics))
        dataList.append(struct.pack(">L", len(self.imageData)))
        dataList.append(self.imageData)
        return bytesjoin(dataList)


class cbdt_bitmap_format_18(BitmapPlusBigMetricsMixin, ColorBitmapGlyph):
    def decompile(self):
        self.metrics = BigGlyphMetrics()
        dummy, data = sstruct.unpack2(bigGlyphMetricsFormat, self.data, self.metrics)
        (dataLen,) = struct.unpack(">L", data[:4])
        data = data[4:]

        # For the image data cut it to the size specified by dataLen.
        assert dataLen <= len(data), "Data overun in format 18"
        self.imageData = data[:dataLen]

    def compile(self, ttFont):
        dataList = []
        dataList.append(sstruct.pack(bigGlyphMetricsFormat, self.metrics))
        dataList.append(struct.pack(">L", len(self.imageData)))
        dataList.append(self.imageData)
        return bytesjoin(dataList)


class cbdt_bitmap_format_19(ColorBitmapGlyph):
    def decompile(self):
        (dataLen,) = struct.unpack(">L", self.data[:4])
        data = self.data[4:]

        assert dataLen <= len(data), "Data overun in format 19"
        self.imageData = data[:dataLen]

    def compile(self, ttFont):
        return struct.pack(">L", len(self.imageData)) + self.imageData


# Dict for CBDT extended formats.
cbdt_bitmap_classes = {
    17: cbdt_bitmap_format_17,
    18: cbdt_bitmap_format_18,
    19: cbdt_bitmap_format_19,
}



import unittest

try:
    from fontTools.pens.quartzPen import QuartzPen

    from Quartz.CoreGraphics import CGPathApply
    from Quartz.CoreGraphics import kCGPathElementMoveToPoint
    from Quartz.CoreGraphics import kCGPathElementAddLineToPoint
    from Quartz.CoreGraphics import kCGPathElementAddQuadCurveToPoint
    from Quartz.CoreGraphics import kCGPathElementAddCurveToPoint
    from Quartz.CoreGraphics import kCGPathElementCloseSubpath

    PATH_ELEMENTS = {
        # CG constant key                    desc       num_points
        kCGPathElementMoveToPoint: ("moveto", 1),
        kCGPathElementAddLineToPoint: ("lineto", 1),
        kCGPathElementAddCurveToPoint: ("curveto", 3),
        kCGPathElementAddQuadCurveToPoint: ("qcurveto", 2),
        kCGPathElementCloseSubpath: ("close", 0),
    }

    PYOBJC_AVAILABLE = True
except ImportError:
    PYOBJC_AVAILABLE = False


def draw(pen):
    pen.moveTo((50, 0))
    pen.lineTo((50, 500))
    pen.lineTo((200, 500))
    pen.curveTo((350, 500), (450, 400), (450, 250))
    pen.curveTo((450, 100), (350, 0), (200, 0))
    pen.closePath()


def quartzPathApplier(elements, element):
    num_points = 0
    elem_type = None
    if element.type in PATH_ELEMENTS:
        num_points = PATH_ELEMENTS[element.type][1]
        elem_type = PATH_ELEMENTS[element.type][0]
    elements.append((elem_type, element.points.as_tuple(num_points)))


def quartzPathElements(path):
    elements = []
    CGPathApply(path, elements, quartzPathApplier)
    return elements


def quartzPathToString(path):
    elements = quartzPathElements(path)
    output = []
    for element in elements:
        elem_type, elem_points = element
        path_points = " ".join([f"{p.x} {p.y}" for p in elem_points])
        output.append(f"{elem_type} {path_points}")
    return " ".join(output)


@unittest.skipUnless(PYOBJC_AVAILABLE, "pyobjc not installed")
class QuartzPenTest(unittest.TestCase):
    def test_draw(self):
        pen = QuartzPen(None)
        draw(pen)
        self.assertEqual(
            "moveto 50.0 0.0 lineto 50.0 500.0 lineto 200.0 500.0 curveto 350.0 500.0 450.0 400.0 450.0 250.0 curveto 450.0 100.0 350.0 0.0 200.0 0.0 close ",
            quartzPathToString(pen.path),
        )

    def test_empty(self):
        pen = QuartzPen(None)
        self.assertEqual("", quartzPathToString(pen.path))


if __name__ == "__main__":
    import sys

    sys.exit(unittest.main())



from fontTools.misc.loggingTools import (
    LevelFormatter,
    Timer,
    configLogger,
    ChannelsFilter,
    LogMixin,
)
from io import StringIO
import logging
import textwrap
import time
import re
import pytest


def logger_name_generator():
    basename = "fontTools.test#"
    num = 1
    while True:
        yield basename + str(num)
        num += 1


unique_logger_name = logger_name_generator()


@pytest.fixture
def logger():
    log = logging.getLogger(next(unique_logger_name))
    configLogger(logger=log, level="DEBUG", stream=StringIO())
    return log


def test_LevelFormatter():
    stream = StringIO()
    handler = logging.StreamHandler(stream)
    formatter = LevelFormatter(
        fmt={
            "*": "[%(levelname)s] %(message)s",
            "DEBUG": "%(name)s [%(levelname)s] %(message)s",
            "INFO": "%(message)s",
        }
    )
    handler.setFormatter(formatter)
    name = next(unique_logger_name)
    log = logging.getLogger(name)
    log.setLevel(logging.DEBUG)
    log.addHandler(handler)

    log.debug("this uses a custom format string")
    log.info("this also uses a custom format string")
    log.warning("this one uses the default format string")

    assert stream.getvalue() == textwrap.dedent(
        """\
        %s [DEBUG] this uses a custom format string
        this also uses a custom format string
        [WARNING] this one uses the default format string
        """
        % name
    )


class TimerTest(object):
    def test_split(self):
        timer = Timer()
        time.sleep(0.01)
        fist_lap = timer.split()
        assert timer.elapsed == fist_lap
        time.sleep(0.1)
        second_lap = timer.split()
        assert second_lap > fist_lap
        assert timer.elapsed == second_lap

    def test_time(self):
        timer = Timer()
        time.sleep(0.01)
        overall_time = timer.time()
        assert overall_time > 0

    def test_context_manager(self):
        with Timer() as t:
            time.sleep(0.01)
        assert t.elapsed > 0

    def test_using_logger(self, logger):
        with Timer(logger, "do something"):
            time.sleep(0.01)

        assert re.match(
            r"Took [0-9]\.[0-9]{3}s to do something",
            logger.handlers[0].stream.getvalue(),
        )

    def test_using_logger_calling_instance(self, logger):
        timer = Timer(logger)
        with timer():
            time.sleep(0.01)

        assert re.match(
            r"elapsed time: [0-9]\.[0-9]{3}s", logger.handlers[0].stream.getvalue()
        )

        # do it again but with custom level
        with timer("redo it", level=logging.WARNING):
            time.sleep(0.02)

        assert re.search(
            r"WARNING: Took [0-9]\.[0-9]{3}s to redo it",
            logger.handlers[0].stream.getvalue(),
        )

    def test_function_decorator(self, logger):
        timer = Timer(logger)

        @timer()
        def test1():
            time.sleep(0.01)

        @timer("run test 2", level=logging.INFO)
        def test2():
            time.sleep(0.02)

        test1()

        assert re.match(
            r"Took [0-9]\.[0-9]{3}s to run 'test1'",
            logger.handlers[0].stream.getvalue(),
        )

        test2()

        assert re.search(
            r"Took [0-9]\.[0-9]{3}s to run test 2", logger.handlers[0].stream.getvalue()
        )


def test_ChannelsFilter(logger):
    n = logger.name
    filtr = ChannelsFilter(n + ".A.B", n + ".C.D")
    handler = logger.handlers[0]
    handler.addFilter(filtr)
    stream = handler.stream

    logging.getLogger(n + ".A.B").debug("this record passes through")
    assert "this record passes through" in stream.getvalue()

    logging.getLogger(n + ".A.B.C").debug("records from children also pass")
    assert "records from children also pass" in stream.getvalue()

    logging.getLogger(n + ".C.D").debug("this one as well")
    assert "this one as well" in stream.getvalue()

    logging.getLogger(n + ".A.B.").debug("also this one")
    assert "also this one" in stream.getvalue()

    before = stream.getvalue()
    logging.getLogger(n + ".A.F").debug("but this one does not!")
    assert before == stream.getvalue()

    logging.getLogger(n + ".C.DE").debug("neither this one!")
    assert before == stream.getvalue()


def test_LogMixin():
    class Base(object):
        pass

    class A(LogMixin, Base):
        pass

    class B(A):
        pass

    a = A()
    b = B()

    assert hasattr(a, "log")
    assert hasattr(b, "log")
    assert isinstance(a.log, logging.Logger)
    assert isinstance(b.log, logging.Logger)
    assert a.log.name == "loggingTools_test.A"
    assert b.log.name == "loggingTools_test.B"



"""Private utility methods used by the subset modules"""


def _add_method(*clazzes):
    """Returns a decorator function that adds a new method to one or
    more classes."""

    def wrapper(method):
        done = []
        for clazz in clazzes:
            if clazz in done:
                continue  # Support multiple names of a clazz
            done.append(clazz)
            assert clazz.__name__ != "DefaultTable", "Oops, table class not found."
            assert not hasattr(
                clazz, method.__name__
            ), "Oops, class '%s' has method '%s'." % (clazz.__name__, method.__name__)
            setattr(clazz, method.__name__, method)
        return None

    return wrapper


def _uniq_sort(l):
    return sorted(set(l))
